# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module for configuring provider credentials and authentication.

"""
from __future__ import annotations

import json
import os
from typing import TYPE_CHECKING, Optional

import boto3
from boto3.session import Session
from braket.ahs.analog_hamiltonian_simulation import AnalogHamiltonianSimulation
from braket.aws import AwsDevice, AwsSession
from braket.circuits import Circuit
from qbraid_core.services.quantum import quantum_lib_proxy_state
from qbraid_core.services.quantum.proxy_braket import aws_configure

from qbraid._caching import cached_method
from qbraid.exceptions import QbraidError
from qbraid.programs import ExperimentType, ProgramSpec
from qbraid.runtime import QuantumProvider, TargetProfile

from .device import BraketDevice

if TYPE_CHECKING:
    import braket.aws

    import qbraid.runtime.aws


class BraketProvider(QuantumProvider):
    """
    This class is responsible for managing the interactions and
    authentications with the AWS services.

    Attributes:
        aws_access_key_id (str): AWS access key ID for authenticating with AWS services.
        aws_secret_access_key (str): AWS secret access key for authenticating with AWS services.
        aws_session_token (str): AWS session token for authenticating with AWS services when using
            temporary credentials.
    """

    REGIONS = ("us-east-1", "us-west-1", "us-west-2", "eu-west-2")

    def __init__(
        self,
        aws_access_key_id: Optional[str] = None,
        aws_secret_access_key: Optional[str] = None,
        aws_session_token: Optional[str] = None,
    ):
        """
        Initializes the QbraidProvider object with optional AWS credentials.

        Args:
            aws_access_key_id (str, optional): AWS access key ID. Defaults to None.
            aws_secret_access_key (str, optional): AWS secret access token. Defaults to None.
            aws_session_token (str, optional): AWS session token. Defaults to None.
        """
        self.aws_access_key_id = aws_access_key_id or os.getenv("AWS_ACCESS_KEY_ID")
        self.aws_secret_access_key = aws_secret_access_key or os.getenv("AWS_SECRET_ACCESS_KEY")
        self.aws_session_token = aws_session_token

    def save_config(
        self,
        aws_access_key_id: Optional[str] = None,
        aws_secret_access_key: Optional[str] = None,
        **kwargs,
    ):
        """Save the current configuration."""
        aws_access_key_id = aws_access_key_id or self.aws_access_key_id
        aws_secret_access_key = aws_secret_access_key or self.aws_secret_access_key
        aws_configure(
            aws_access_key_id=aws_access_key_id,
            aws_secret_access_key=aws_secret_access_key,
            **kwargs,
        )

    @staticmethod
    def _get_default_region() -> str:
        """Returns the default AWS region."""
        default_region = "us-east-1"

        try:
            session = Session()
            return session.region_name or default_region
        except Exception:  # pylint: disable=broad-exception-caught
            return default_region

    @staticmethod
    def _get_s3_default_bucket() -> str:
        """Return name of the default bucket to use in the AWS Session"""
        try:
            aws_session = AwsSession()
            return aws_session.default_bucket()
        except Exception:  # pylint: disable=broad-exception-caught
            return "amazon-braket-qbraid-jobs"

    def _get_aws_session(self, region_name: Optional[str] = None) -> braket.aws.AwsSession:
        """Returns the AwsSession provider."""
        region_name = region_name or self._get_default_region()
        default_bucket = self._get_s3_default_bucket()

        boto_session = Session(
            region_name=region_name,
            aws_access_key_id=self.aws_access_key_id,
            aws_secret_access_key=self.aws_secret_access_key,
            aws_session_token=self.aws_session_token,
        )
        braket_client = boto_session.client("braket", region_name=region_name)
        return AwsSession(
            boto_session=boto_session, braket_client=braket_client, default_bucket=default_bucket
        )

    def _build_runtime_profile(
        self, device: braket.aws.AwsDevice, program_spec: Optional[ProgramSpec] = None, **kwargs
    ) -> TargetProfile:
        """Returns the runtime profile for the device."""
        metadata = device.aws_session.get_device(device.arn)
        simulator = metadata.get("deviceType") == "SIMULATOR"
        provider_name = metadata.get("providerName")
        capabilities: dict = json.loads(metadata.get("deviceCapabilities"))
        paradigm: dict = capabilities.get("paradigm", {})
        action: dict = capabilities.get("action", {})
        num_qubits = paradigm.get("qubitCount")
        basis_gates = paradigm.get("nativeGateSet", None)
        if action.get("braket.ir.openqasm.program") is not None:
            experiment_type = ExperimentType.GATE_MODEL
            program_spec = program_spec or ProgramSpec(Circuit)
        elif action.get("braket.ir.ahs.program") is not None:
            experiment_type = ExperimentType.AHS
            program_spec = program_spec or ProgramSpec(
                AnalogHamiltonianSimulation, alias="braket_ahs"
            )
        else:
            raise QbraidError(
                f"TargetProfile cannot be created for device '{device.arn}' as it does not "
                "support 'braket.ir.openqasm.program' or 'braket.ir.ahs.program' types. "
                "Please verify device capabilities or select a different, compatible device."
            )
        return TargetProfile(
            simulator=simulator,
            num_qubits=num_qubits,
            experiment_type=experiment_type,
            program_spec=program_spec,
            provider_name=provider_name,
            device_id=device.arn,
            basis_gates=basis_gates,
            **kwargs,
        )

    @cached_method
    def get_devices(
        self,
        aws_session: Optional[braket.aws.AwsSession] = None,
        statuses: Optional[list[str]] = None,
        **kwargs,
    ) -> list[qbraid.runtime.aws.BraketDevice]:
        """Return a list of backends matching the specified filtering."""
        aws_session = self._get_aws_session() if aws_session is None else aws_session
        statuses = ["ONLINE", "OFFLINE"] if statuses is None else statuses
        aws_devices = AwsDevice.get_devices(aws_session=aws_session, statuses=statuses, **kwargs)
        return [
            BraketDevice(profile=self._build_runtime_profile(device), session=device.aws_session)
            for device in aws_devices
        ]

    @cached_method
    def get_device(
        self,
        device_id: str,
    ) -> qbraid.runtime.aws.BraketDevice:
        """Returns the AWS device."""
        try:
            region_name = device_id.split(":")[3]
        except IndexError as err:
            raise ValueError(
                f"Device ARN is not a valid format: {device_id}. For valid Braket ARNs, "
                "see 'https://docs.aws.amazon.com/braket/latest/developerguide/braket-devices.html'"
            ) from err
        region_name = region_name or self._get_default_region()
        aws_session = self._get_aws_session(region_name=region_name)
        device = AwsDevice(arn=device_id, aws_session=aws_session)
        profile = self._build_runtime_profile(device)
        return BraketDevice(profile=profile, session=device.aws_session)

    @staticmethod
    def _fetch_resources(region_names: list[str], key: str, values: list[str]) -> list[str]:
        """Fetch resources from AWS."""
        tasks = []
        for region_name in region_names:
            client = boto3.client("resourcegroupstaggingapi", region_name=region_name)
            response = client.get_resources(
                TagFilters=[
                    {
                        "Key": key,
                        "Values": values,
                    }
                ],
            )
            matches = [t["ResourceARN"] for t in response["ResourceTagMappingList"]]
            tasks += matches
        return tasks

    def get_tasks_by_tag(
        self, key: str, values: Optional[list[str]] = None, region_names: Optional[list[str]] = None
    ) -> list[str]:
        """
        Retrieves a list of quantum task ARNs that match the specified tag keys or key/value pairs.

        Args:
            key (str): The tag key to match.
            values (Optional[list[str]]): A list of tag values to match against the provided
                                          key. If None, tasks with the specified key,
                                          regardless of its value, are matched.
            region_names (Optional[list[str]]): A list of region names to search. If None, all
                                                regions in `self.REGIONS` are searched.

        Returns:
            list[str]: A list of ARNs for quantum tasks that match the given tag criteria.

        Raises:
            QbraidError: If the function is called within a qBraid quantum job environment
                         where AWS S3 requests are not supported.
        """
        try:
            jobs_state = quantum_lib_proxy_state("braket")
            jobs_enabled = jobs_state.get("enabled", False)
        except ValueError:
            jobs_enabled = False

        if jobs_enabled:
            raise QbraidError("AWS S3 requests not supported by qBraid quantum jobs.")

        region_names = (
            region_names
            if region_names is not None and len(region_names) > 0
            else list(self.REGIONS)
        )
        values = values if values is not None else []

        return self._fetch_resources(region_names, key, values)

    def __hash__(self):
        if not hasattr(self, "_hash"):
            object.__setattr__(
                self,
                "_hash",
                hash((self.aws_access_key_id, self.aws_secret_access_key, self.aws_session_token)),
            )
        return self._hash  # pylint: disable=no-member



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Mdule submiting and managing jobs through the IonQ API.

.. currentmodule:: qbraid.runtime.ionq

Classes
--------

.. autosummary::
   :toctree: ../stubs/

    IonQSession
    IonQProvider
    IonQDevice
    IonQJob

Exceptions
-----------

.. autosummary::
   :toctree: ../stubs/

    IonQJobError

"""
from .device import IonQDevice
from .job import IonQJob, IonQJobError
from .provider import IonQProvider, IonQSession

__all__ = [
    "IonQDevice",
    "IonQProvider",
    "IonQSession",
    "IonQJob",
    "IonQJobError",
]



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module containing Qiskit quantum circuits used for testing

"""
import numpy as np
from qiskit import QuantumCircuit


def qiskit_bell() -> QuantumCircuit:
    """Returns Qiskit bell circuit"""
    circuit = QuantumCircuit(2)
    circuit.h(0)
    circuit.cx(0, 1)
    return circuit


def qiskit_shared15():
    """Returns qiskit `QuantumCircuit` for qBraid `TestSharedGates`."""

    circuit = QuantumCircuit(4)

    circuit.h([0, 1, 2, 3])
    circuit.x([0, 1])
    circuit.y(2)
    circuit.z(3)
    circuit.s(0)
    circuit.sdg(1)
    circuit.t(2)
    circuit.tdg(3)
    circuit.rx(np.pi / 4, 0)
    circuit.ry(np.pi / 2, 1)
    circuit.rz(3 * np.pi / 4, 2)
    circuit.p(np.pi / 8, 3)
    circuit.sx(0)
    circuit.sxdg(1)
    circuit.iswap(2, 3)
    circuit.swap([0, 1], [2, 3])
    circuit.cx(0, 1)
    circuit.cp(np.pi / 4, 2, 3)

    return circuit



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module defining all :mod:`~qbraid.runtime` enumerated types.

"""
from __future__ import annotations

from enum import Enum, IntEnum


class DeviceStatus(Enum):
    """Enumeration for representing various operational statuses of devices.

    Attributes:
        ONLINE (str): Device is online and accepting jobs.
        UNAVAILABLE (str): Device is online but not accepting jobs.
        OFFLINE (str): Device is offline.
        RETIRED (str): Device has been retired and is no longer operational.
    """

    ONLINE = "online"
    UNAVAILABLE = "unavailable"
    OFFLINE = "offline"
    RETIRED = "retired"


class JobStatus(Enum):
    """Enum for the status of processes (i.e. quantum jobs / tasks) resulting
    from any :meth:`~qbraid.runtime.QuantumDevice.run` method.

    Displayed status text values may differ from those listed below to provide
    additional visibility into tracebacks, particularly for failed jobs.

    """

    def __new__(cls, value: str):
        """Enumeration representing the status of a :py:class:`QuantumJob`."""
        obj = object.__new__(cls)
        obj._value_ = value
        obj.default_message = cls._get_default_message(value)
        obj.status_message = None
        return obj

    @classmethod
    def _get_default_message(cls, status: str) -> str:
        """Get the default message for a given status value."""
        default_messages = {
            "INITIALIZING": "job is being initialized",
            "QUEUED": "job is queued",
            "VALIDATING": "job is being validated",
            "RUNNING": "job is actively running",
            "CANCELLING": "job is being cancelled",
            "CANCELLED": "job has been cancelled",
            "COMPLETED": "job has successfully run",
            "FAILED": "job failed / incurred error",
            "UNKNOWN": "job status is unknown/undetermined",
            "HOLD": "job terminal but results withheld due to account status",
        }
        message = default_messages.get(status)

        if message is None:
            raise ValueError(f"Invalid status value: {status}")

        return message

    def set_status_message(self, message: str) -> None:
        """Set a custom message for the enum instance."""
        self.status_message = message

    def __repr__(self):
        """Custom repr to show custom message or default."""
        message = self.status_message if self.status_message else self.default_message
        return f"<{self.name}: '{message}'>"

    def __call__(self) -> JobStatus:
        """Create a new instance of the enum member, allowing unique attributes."""
        obj = self.__class__(self._value_)
        obj.default_message = self.default_message
        return obj

    @classmethod
    def terminal_states(cls) -> set[JobStatus]:
        """Returns the final job statuses."""
        return {cls.COMPLETED, cls.CANCELLED, cls.FAILED}

    INITIALIZING = "INITIALIZING"
    QUEUED = "QUEUED"
    VALIDATING = "VALIDATING"
    RUNNING = "RUNNING"
    CANCELLING = "CANCELLING"
    CANCELLED = "CANCELLED"
    COMPLETED = "COMPLETED"
    FAILED = "FAILED"
    UNKNOWN = "UNKNOWN"
    HOLD = "HOLD"


class ValidationLevel(IntEnum):
    """Enumeration for program validation levels in qBraid runtime.

    Attributes:
        NONE (int): No validation is performed.
        WARN (int): Validation is performed, and warnings are issued if validation fails.
        RAISE (int): Validation is performed, and exceptions are raised if validation fails.
    """

    NONE = 0
    WARN = 1
    RAISE = 2



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Unit tests for the qBraid runtime Options dataclass.

"""
import copy

import pytest

from qbraid.runtime.options import RuntimeOptions


def test_options_initialization():
    """Test that the Options class initializes with dynamic fields."""
    options = RuntimeOptions(transpile=True, custom_field=42)
    assert options.get("transpile") is True
    assert options.get("custom_field") == 42


def test_options_set_and_get_field():
    """Test setting and getting a field using both attribute and dictionary access."""
    options = RuntimeOptions()
    options.transpile = False
    options["custom_field"] = "test_value"

    assert options.transpile is False
    assert options["custom_field"] == "test_value"


def test_options_get_with_fallback():
    """Test get method with a fallback value for non-existent fields."""
    options = RuntimeOptions(transpile=True)
    assert options.get("non_existent", "default") == "default"
    assert options.get("transpile") is True


def test_options_setitem_getitem():
    """Test __setitem__ and __getitem__ for setting and accessing fields."""
    options = RuntimeOptions()
    options["new_field"] = 123
    assert options["new_field"] == 123


def test_options_delitem():
    """Test __delitem__ behavior for deleting dynamic fields but not default ones."""
    options = RuntimeOptions(transpile=True)

    options["custom_field"] = 42
    assert options["custom_field"] == 42

    del options["custom_field"]
    with pytest.raises(KeyError):
        _ = options["custom_field"]

    with pytest.raises(KeyError):
        del options["transpile"]


def test_options_getattr_setattr():
    """Test __getattr__ and __setattr__ for setting and getting attributes."""
    options = RuntimeOptions()
    options.custom_attr = 99
    assert options.custom_attr == 99

    with pytest.raises(AttributeError):
        _ = options.non_existent_attr


def test_options_set_validator_and_validation():
    """Test setting a validator and ensuring validation works properly."""
    options = RuntimeOptions(transpile=True)

    options.set_validator("transpile", lambda x: isinstance(x, bool))

    options.transpile = False
    assert options.transpile is False

    with pytest.raises(ValueError):
        options.transpile = "invalid_value"


def test_options_update_options():
    """Test update_options method for updating multiple options with validation."""
    options = RuntimeOptions()

    options.update_options(transpile=False, custom_field=123)
    options.set_validator("custom_field", lambda x: isinstance(x, int))

    assert options.transpile is False
    assert options["custom_field"] == 123

    with pytest.raises(ValueError):
        options.update_options(custom_field="invalid")


def test_options_repr():
    """Test the __repr__ method for correct string output."""
    options = RuntimeOptions(transpile=True, custom_field=42)
    repr_str = repr(options)
    assert repr_str == "RuntimeOptions(transpile=True, custom_field=42)"


def test_options_get_with_missing_field():
    """Test that missing fields return None if no default is provided."""
    options = RuntimeOptions()
    assert options.get("missing_field") is None


def test_options_setattr_update_through_getattr():
    """Test setting a value using setattr and ensuring it reflects via getattr."""
    options = RuntimeOptions()
    setattr(options, "new_field", "test_value")
    assert getattr(options, "new_field") == "test_value"


def test_options_delitem_non_existent_field():
    """Test __delitem__ raises KeyError when deleting a non-existent field."""
    options = RuntimeOptions(transpile=True)
    with pytest.raises(KeyError):
        del options["non_existent_field"]


def test_options_remove_validator():
    """Test that removing a validator allows invalid values."""
    options = RuntimeOptions(transpile=True)

    options.set_validator("transpile", lambda x: isinstance(x, bool))
    with pytest.raises(ValueError):
        options.transpile = "invalid_value"

    del options._validators["transpile"]

    options.transpile = "no_validation_now"
    assert options.transpile == "no_validation_now"


def test_options_multiple_validators():
    """Test that setting multiple validators overwrites the previous one."""
    options = RuntimeOptions(custom_field=10)
    options.set_validator("custom_field", lambda x: isinstance(x, int))

    options.set_validator("custom_field", lambda x: x > 5)

    with pytest.raises(ValueError):
        options.custom_field = 3


def test_options_validator_called_only_on_update():
    """Test that the validator is called only when updating a value."""
    options = RuntimeOptions(transpile=True)
    validation_called = False

    def validator(x):
        nonlocal validation_called
        validation_called = True
        return isinstance(x, bool)

    options.set_validator("transpile", validator)

    options.transpile = False
    assert validation_called is True

    validation_called = False
    options.transpile = False
    assert validation_called is False


def test_options_empty_initialization():
    """Test that the Options class can initialize without arguments."""
    options = RuntimeOptions()
    assert options._fields == {}
    assert options.get("non_existent") is None


def test_options_none_as_value():
    """Test that None can be set and retrieved as a valid field value."""
    options = RuntimeOptions()
    options.custom_field = None
    assert options.custom_field is None
    assert options.get("custom_field") is None


def test_options_reset_default_field():
    """Test that a default field can be reset."""
    options = RuntimeOptions(transpile=True)
    options.transpile = False
    assert options.transpile is False
    options.transpile = True
    assert options.transpile is True


def test_options_dict_property():
    """Test that the __dict__ property returns the internal fields."""
    options = RuntimeOptions(transpile=True, custom_field=42)
    assert options.__dict__ == {"transpile": True, "custom_field": 42}


def test_options_dict_conversion():
    """Test that dict(options) works correctly and returns the internal fields."""
    options = RuntimeOptions(transpile=True, custom_field=42)
    options_dict = dict(options)
    assert options_dict == {"transpile": True, "custom_field": 42}


def test_options_iter_method():
    """Test that __iter__ allows iteration over key-value pairs."""
    options = RuntimeOptions(transpile=True, custom_field=42)

    key_value_pairs = list(iter(options))
    expected_pairs = [("transpile", True), ("custom_field", 42)]

    assert key_value_pairs == expected_pairs


def test_options_len_method():
    """Test that __len__ returns the number of internal fields."""
    options = RuntimeOptions(transpile=True, custom_field=42)
    assert len(options) == 2

    options["new_field"] = "new_value"
    assert len(options) == 3


def test_options_equality_check():
    """Test custom __eq__ method with various edge cases."""
    op1 = RuntimeOptions(x=2)
    op2 = RuntimeOptions(x=2)
    assert op1 == op2

    op1.hello = "world"
    assert op1 != op2

    op2.hello = "world"
    assert op1 == op2

    op1.set_validator("hello", lambda x: isinstance(x, str))
    assert op1 != op2

    op2.set_validator("hello", lambda x: isinstance(x, str))
    assert op1 == op2


def test_options_copy_creates_new_instance():
    """Test that __copy__ creates a new instance of Options."""
    options = RuntimeOptions(transpile=True, custom_field=42)
    copied_options = copy.copy(options)

    assert copied_options is not options
    assert isinstance(copied_options, RuntimeOptions)


def test_options_copy_has_same_fields():
    """Test that the copied instance has the same fields as the original."""
    options = RuntimeOptions(transpile=True, custom_field=42)
    copied_options = copy.copy(options)

    assert copied_options._fields == options._fields
    assert copied_options["transpile"] == options["transpile"]
    assert copied_options["custom_field"] == options["custom_field"]


def test_options_copy_independent_modification():
    """Test that modifying the copied instance doesn't affect the original."""
    options = RuntimeOptions(transpile=True, custom_field=42)
    copied_options = copy.copy(options)

    copied_options["custom_field"] = 100

    assert options["custom_field"] == 42
    assert copied_options["custom_field"] == 100


def test_options_copy_with_validators():
    """Test that validators are copied and work independently on the copy."""
    options = RuntimeOptions(transpile=True)
    options.set_validator("transpile", lambda x: isinstance(x, bool))

    copied_options = copy.copy(options)

    assert "transpile" in copied_options._validators
    assert copied_options._validators["transpile"] is not None

    copied_options.transpile = False
    assert copied_options.transpile is False

    with pytest.raises(ValueError):
        copied_options.transpile = "invalid"

    with pytest.raises(ValueError):
        options.transpile = "invalid"


def test_options_copy_with_dynamic_fields():
    """Test that dynamic fields are copied correctly."""
    options = RuntimeOptions(transpile=True)
    options.dynamic_field = "dynamic_value"

    copied_options = copy.copy(options)

    assert copied_options.dynamic_field == "dynamic_value"

    copied_options.dynamic_field = "modified_value"
    assert copied_options.dynamic_field == "modified_value"
    assert options.dynamic_field == "dynamic_value"


def test_options_reserved_keyword_set_validator():
    """Test that 'set_validator' keyword raises a ValueError."""
    with pytest.raises(
        ValueError, match="The option name 'set_validator' is reserved and cannot be used."
    ):
        RuntimeOptions(set_validator=True)


def test_options_reserved_keyword_validate_option():
    """Test that 'validate_option' keyword raises a ValueError."""
    with pytest.raises(
        ValueError, match="The option name 'validate_option' is reserved and cannot be used."
    ):
        RuntimeOptions(validate_option=True)


def test_options_reserved_keyword_update():
    """Test that 'update_options' keyword raises a ValueError."""
    with pytest.raises(
        ValueError, match="The option name 'update_options' is reserved and cannot be used."
    ):
        RuntimeOptions(update_options=True)


def test_options_reserved_keyword_get():
    """Test that 'get' keyword raises a ValueError."""
    with pytest.raises(ValueError, match="The option name 'get' is reserved and cannot be used."):
        RuntimeOptions(get=True)


def test_options_dunder_keyword():
    """Test that dunder-prefixed keywords raise a ValueError."""
    with pytest.raises(
        ValueError, match="The option name '__hidden' with dunder prefix is not allowed."
    ):
        RuntimeOptions(__hidden=True)


def test_options_valid_keywords():
    """Test that valid keywords do not raise errors."""
    try:
        options = RuntimeOptions(transpile=True, debug=False)
        assert options.transpile is True
        assert options.debug is False
    except ValueError:
        pytest.fail("ValueError raised for valid keywords")


def test_options_equality_type_mismatch():
    """Test that equality check returns False for different types."""
    options = RuntimeOptions(transpile=True)
    assert options != "not_options"


def test_merge_non_overlapping_options():
    """Test merging options with non-overlapping fields."""
    options1 = RuntimeOptions(option_a=1)
    options2 = RuntimeOptions(option_b=2)

    options1.merge(options2)

    assert options1.option_a == 1
    assert options1.option_b == 2


def test_merge_new_option_with_validator():
    """Test merging when 'other' has a new option with a validator."""
    options1 = RuntimeOptions(option_a=1)

    options2 = RuntimeOptions(option_b=2)
    options2.set_validator("option_b", lambda x: isinstance(x, int) and x > 0)

    options1.merge(options2)

    assert options1.option_b == 2

    # Validator for option_b should be in place
    with pytest.raises(ValueError):
        options1.option_b = -1


@pytest.mark.parametrize(
    "override_validators, initial_value1, validator1, value2, validator2, test_value, should_raise",
    [
        (
            True,
            1,
            lambda x: isinstance(x, int) and x > 0,
            2,
            lambda x: isinstance(x, int) and x > 1,
            1,
            True,
        ),
        (
            False,
            1,
            lambda x: isinstance(x, int) and x > 0,
            2,
            lambda x: isinstance(x, int) and x > 1,
            1,
            False,
        ),
    ],
)

# pylint: disable-next=too-many-arguments
def test_merge_overlapping_options(
    override_validators,
    initial_value1,
    validator1,
    value2,
    validator2,
    test_value,
    should_raise,
):
    """Test merging options with overlapping fields and validators."""
    options1 = RuntimeOptions(option_a=initial_value1)
    options1.set_validator("option_a", validator1)

    options2 = RuntimeOptions(option_a=value2)
    options2.set_validator("option_a", validator2)

    # Handle potential merge failure
    if not override_validators and not validator1(value2):
        with pytest.raises(ValueError):
            options1.merge(options2, override_validators=override_validators)
        return

    options1.merge(options2, override_validators=override_validators)

    assert options1.option_a == value2

    if should_raise:
        with pytest.raises(ValueError):
            options1.option_a = test_value
    else:
        options1.option_a = test_value
        assert options1.option_a == test_value


def test_merge_existing_option_invalidated_by_new_validator():
    """Test merging when existing option in 'self' is invalid under new validator from 'other'."""
    options1 = RuntimeOptions(option_a=1)
    options1.set_validator("option_a", lambda x: x == 1)

    options2 = RuntimeOptions()
    options2.set_validator("option_a", lambda x: x == 2)

    options2.merge(options1, override_validators=True)

    with pytest.raises(ValueError):
        options2.option_a = 2


def test_merge_preserve_existing_validators():
    """Test merging with override_validators=False to preserve existing validators."""
    options1 = RuntimeOptions(option_a=2)
    options1.set_validator("option_a", lambda x: isinstance(x, int) and x % 2 == 0)

    options2 = RuntimeOptions(option_a=4)  # Even number
    options2.set_validator("option_a", lambda x: isinstance(x, int) and x > 0)

    options1.merge(options2, override_validators=False)

    # Validator from options1 should be preserved
    with pytest.raises(ValueError):
        options1.option_a = 3

    options1.option_a = 6
    assert options1.option_a == 6


def test_merge_override_validators():
    """Test merging with override_validators=True to override existing validators."""
    options1 = RuntimeOptions(option_a=2)
    options1.set_validator("option_a", lambda x: isinstance(x, int) and x % 2 == 0)

    options2 = RuntimeOptions(option_a=3)
    options2.set_validator("option_a", lambda x: isinstance(x, int) and x > 0)

    options1.merge(options2, override_validators=True)

    options1.option_a = 5
    assert options1.option_a == 5

    with pytest.raises(ValueError):
        options1.option_a = -1


def test_merge_with_existing_option_invalid_under_new_validator():
    """Test merging when existing option value is invalid under new validator."""
    options1 = RuntimeOptions(option_a=4)

    options2 = RuntimeOptions(option_a=3)
    options2.set_validator("option_a", lambda x: x % 2 != 0)  # Odd numbers only

    with pytest.raises(
        ValueError, match="Value '4' is not valid for field 'option_a' after merging."
    ):
        options2.merge(options1)


def test_merge_with_no_validators():
    """Test merging when neither 'self' nor 'other' have validators."""
    options1 = RuntimeOptions(option_a="value1")
    options2 = RuntimeOptions(option_b="value2")

    options1.merge(options2)

    assert options1.option_a == "value1"
    assert options1.option_b == "value2"


def test_merge_with_invalid_option_in_other():
    """Test merging when 'other' has an invalid option value according to its own validator."""
    options1 = RuntimeOptions(option_a=1)
    options1.set_validator("option_b", lambda x: x == "valid")

    options2 = RuntimeOptions(option_b="invalid")

    with pytest.raises(
        ValueError, match="Value 'invalid' is not valid for field 'option_b' after merging."
    ):
        options1.merge(options2)


def test_merge_with_multiple_options_and_validators():
    """Test merging multiple options and validators."""
    options1 = RuntimeOptions(option_a=1, option_c=3)
    options1.set_validator("option_a", lambda x: isinstance(x, int))
    options1.set_validator("option_c", lambda x: x in [3, 4, 5])

    options2 = RuntimeOptions(option_b="test", option_c=4)
    options2.set_validator("option_b", lambda x: isinstance(x, str))
    options2.set_validator("option_c", lambda x: x in [4, 5, 6])

    options1.merge(options2, override_validators=True)

    assert options1.option_a == 1
    assert options1.option_b == "test"
    assert options1.option_c == 4

    # Validator from options2 should be in effect for option_c
    with pytest.raises(ValueError):
        options1.option_c = 3  # 3 not in [4, 5, 6]

    options1.option_c = 5  # Should pass
    assert options1.option_c == 5


def test_validate_option_handles_exception():
    """Test that validate_option handles exceptions raised by the validator."""
    options = RuntimeOptions(v=2)
    options.set_validator("v", lambda x: x > 0)
    with pytest.raises(ValueError) as exc:
        options.validate_option("v", "a")
    assert (
        "Validator for field 'v' raised an exception: "
        "'>' not supported between instances of 'str' and 'int'" == str(exc.value)
    )


def test_options_set_validator_for_existing_field():
    """Test setting a validator for an existing field."""
    options = RuntimeOptions(test=True)
    options.set_validator("test", lambda x: isinstance(x, bool))

    assert options.test is True

    with pytest.raises(ValueError):
        options.set_validator("test", lambda x: not isinstance(x, bool))


def test_options_set_validator_for_existing_field_with_invalid_value():
    """Test setting a validator for an existing field with an invalid value."""
    options = RuntimeOptions(custom_field="string")

    with pytest.raises(
        ValueError,
        match="Existing value 'string' for field 'custom_field' is not valid for the new validator",
    ):
        options.set_validator("custom_field", lambda x: isinstance(x, int))


def test_options_set_validator_for_non_existent_field():
    """Test setting a validator for a non-existent field."""
    options = RuntimeOptions()
    options.set_validator("new_field", lambda x: isinstance(x, int))

    assert "new_field" in options._validators


def test_options_validator_applies_to_existing_field():
    """Test that a validator applies to an existing field."""
    options = RuntimeOptions(custom_field=42)
    options.set_validator("custom_field", lambda x: isinstance(x, int) and x > 0)

    assert options.custom_field == 42

    with pytest.raises(ValueError):
        options.custom_field = -1


def test_options_validator_allows_valid_update_to_existing_field():
    """Test that a validator allows a valid update to an existing field."""
    options = RuntimeOptions(custom_field=42)
    options.set_validator("custom_field", lambda x: isinstance(x, int) and x > 0)

    options.custom_field = 100
    assert options.custom_field == 100


def test_options_set_validator_error_message_for_exception():
    """Test that the error message for a validator exception is correct."""
    options = RuntimeOptions(custom_field="string")

    with pytest.raises(
        ValueError,
        match=(
            "Existing value 'string' for field 'custom_field' "
            "raised an exception against the new validator"
        ),
    ):
        options.set_validator("custom_field", lambda x: x > 0)


def test_options_set_validator_error_message_for_invalid_value():
    """Test that the error message for an invalid value is correct."""
    options = RuntimeOptions(custom_field=-5)

    with pytest.raises(
        ValueError,
        match="Existing value '-5' for field 'custom_field' is not valid for the new validator",
    ):
        options.set_validator("custom_field", lambda x: x > 0)


def test_options_set_validator_error_message_with_existing_validator():
    """Test that the error message for an existing validator is correct."""
    options = RuntimeOptions(custom_field=5)
    options.set_validator("custom_field", lambda x: x > 0)

    with pytest.raises(
        ValueError, match="Please delete the field before setting the new validator"
    ):
        options.set_validator("custom_field", lambda x: x < 0)



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module defining IonQ device class

"""
from __future__ import annotations

import importlib.util
import json
import warnings
from typing import TYPE_CHECKING, Any, Literal, Optional, Union

import pyqasm
from qbraid_core._import import LazyLoader

from qbraid._logging import logger
from qbraid.passes import CompilationError
from qbraid.programs import QPROGRAM_REGISTRY, load_program
from qbraid.programs.gate_model.ionq import GateSet, InputFormat
from qbraid.programs.gate_model.qasm2 import OpenQasm2Program
from qbraid.programs.gate_model.qasm3 import OpenQasm3Program
from qbraid.programs.typer import IonQDict, IonQDictType, QasmStringType
from qbraid.runtime.device import QuantumDevice
from qbraid.runtime.enums import DeviceStatus
from qbraid.transpiler.conversions.openqasm3.openqasm3_to_ionq import (
    IONQ_ONE_QUBIT_GATE_MAP,
    IONQ_THREE_QUBIT_GATE_ALIASES,
    IONQ_TWO_QUBIT_GATE_MAP,
)

from .job import IonQJob

if TYPE_CHECKING:
    import qiskit as qiskit_typing

    import qbraid.runtime
    import qbraid.runtime.ionq.provider

qiskit = LazyLoader("qiskit", globals(), "qiskit")
qiskit_ionq = LazyLoader("qiskit_ionq", globals(), "qiskit_ionq")

IONQ_GATE_MAP = IONQ_ONE_QUBIT_GATE_MAP | IONQ_TWO_QUBIT_GATE_MAP | IONQ_THREE_QUBIT_GATE_ALIASES

DEFAULT_FORMAT = InputFormat.CIRCUIT.value
DEFAULT_GATESET = GateSet.QIS.value


class IonQDevice(QuantumDevice):
    """IonQ quantum device interface."""

    def __init__(
        self,
        profile: qbraid.runtime.TargetProfile,
        session: qbraid.runtime.ionq.provider.IonQSession,
    ):
        super().__init__(profile=profile)
        self._session = session

    @property
    def session(self) -> qbraid.runtime.ionq.provider.IonQSession:
        """Return the IonQ session."""
        return self._session

    def __str__(self):
        """String representation of the IonQDevice object."""
        return f"{self.__class__.__name__}('{self.id}')"

    def status(self) -> qbraid.runtime.DeviceStatus:
        """Return the current status of the IonQ device."""
        device_data = self.session.get_device(self.id)
        status = device_data.get("status")

        if status in ["available", "running"]:
            return DeviceStatus.ONLINE

        if status in ["unavailable", "reserved", "calibrating"]:
            return DeviceStatus.UNAVAILABLE

        if status == "retired":
            return DeviceStatus.RETIRED

        if status == "offline":
            return DeviceStatus.OFFLINE

        raise ValueError(f"Unrecognized device status: {status}")

    def avg_queue_time(self) -> int:
        """Return the average queue time for the IonQ device."""
        device_data = self.session.get_device(self.id)
        return device_data["average_queue_time"]

    def transform(self, run_input: QasmStringType) -> QasmStringType:
        """Transform the input to the IonQ device."""
        program: OpenQasm2Program | OpenQasm3Program = load_program(run_input)

        try:
            program.transform(device=self, gate_mappings=IONQ_GATE_MAP)
        except CompilationError as err:
            logger.debug("Failed to transform OpenQASM program for IonQ: %s", err)
            logger.debug("Retrying using pyqasm.unroll()...")
            program._module.unroll()
            program._program = pyqasm.dumps(program._module)
            program.transform(device=self, gate_mappings=IONQ_GATE_MAP)

        return program.program

    @staticmethod
    def _squash_multicircuit_input(batch_input: list[IonQDictType]) -> dict[str, Any]:
        if not batch_input:
            raise ValueError("run_input list cannot be empty.")

        input_format = batch_input[0].get("format", DEFAULT_FORMAT)
        input_gateset = batch_input[0].get("gateset", DEFAULT_GATESET)
        max_qubits = 0
        circuits = []

        for i, run_input in enumerate(batch_input):
            if not isinstance(run_input, IonQDict):
                raise ValueError("All run_inputs must be an instance of ~IonQDict.")
            if run_input.get("format", DEFAULT_FORMAT) != input_format:
                raise ValueError("All run_inputs must have the same value for key 'format'.")
            if run_input.get("gateset", DEFAULT_FORMAT) != input_gateset:
                raise ValueError("All run_inputs must have the same value for key 'gateset'.")

            max_qubits = max(max_qubits, run_input["qubits"])
            circuits.append({"circuit": run_input["circuit"], "name": f"Circuit {i}"})

        return {
            "format": input_format,
            "gateset": input_gateset,
            "qubits": max_qubits,
            "circuits": circuits,
        }

    # pylint:disable-next=arguments-differ,too-many-arguments
    def submit(
        self,
        run_input: Union[IonQDictType, list[IonQDictType]],
        shots: int,
        preflight: bool = False,
        name: Optional[str] = None,
        noise: Optional[dict[str, Any]] = None,
        error_mitigation: Optional[dict[str, Any]] = None,
        metadata: Optional[dict[str, Any]] = None,
        **kwargs,
    ) -> IonQJob:
        """Submit a job to the IonQ device."""
        ionq_input = (
            self._squash_multicircuit_input(run_input) if isinstance(run_input, list) else run_input
        )
        job_data = {
            "target": self.id,
            "shots": shots,
            "preflight": preflight,
            "input": ionq_input,
            **kwargs,
        }
        optional_fields = {
            "name": name,
            "noise": noise,
            "metadata": metadata,
            "error_mitigation": error_mitigation,
        }
        job_data.update({key: value for key, value in optional_fields.items() if value is not None})
        serialized_data = json.dumps(job_data)
        job_data = self.session.create_job(serialized_data)
        job_id = job_data.get("id")
        if not job_id:
            raise ValueError("Job ID not found in the response")
        return IonQJob(job_id=job_id, session=self.session, device=self, shots=shots)

    def _apply_qiskit_ionq_conversion(
        self,
        run_input: list[qiskit_typing.QuantumCircuit],
        gateset: Literal["qis", "native"] = "qis",
        ionq_compiler_synthesis: bool = False,
    ) -> list[IonQDictType]:
        # pylint: disable-next=import-outside-toplevel
        from qbraid.transpiler.conversions.qiskit import qiskit_to_ionq

        provider = qiskit_ionq.IonQProvider(token=self.session.api_key)
        backend = provider.get_backend(self.id, gateset=gateset)

        run_input_compat = []
        for program in run_input:
            transpiled_circuit = qiskit.transpile(program, backend=backend)
            ionq_dict = qiskit_to_ionq(
                transpiled_circuit, gateset=gateset, ionq_compiler_synthesis=ionq_compiler_synthesis
            )
            run_input_compat.append(ionq_dict)

        return run_input_compat

    def run(
        self,
        run_input: Union[qbraid.programs.QPROGRAM, list[qbraid.programs.QPROGRAM]],
        *args,
        gateset: Optional[GateSet] = None,
        ionq_compiler_synthesis: Optional[bool] = None,
        **kwargs,
    ) -> Union[qbraid.runtime.IonQJob, list[qbraid.runtime.IonQJob]]:
        """
        Run a quantum job or a list of quantum jobs on this IonQ device.

        Args:
            run_input: A single quantum program or a list of quantum programs to run on the device.
            gateset (GateSet, optional): The gate set to use for the qiskit-ionq transpile step.
                Only applicable if qiskit-ionq is installed and all run_inputs are of type
                qiskit.QuantumCircuit. Defaults to GateSet.QIS.
            ionq_compiler_synthesis (bool, optional): Whether to opt-in to IonQ compiler's
                intelligent trotterization. Only applicable if qiskit-ionq is installed and all
                run_inputs are of type qiskit.QuantumCircuit. Defaults to False.

        Returns:
            An IonQJob object or a list of IonQJob objects corresponding to the input.
        """
        is_single_input = not isinstance(run_input, list)
        run_input = [run_input] if is_single_input else run_input

        if (
            "qiskit" in QPROGRAM_REGISTRY
            and all(isinstance(program, QPROGRAM_REGISTRY["qiskit"]) for program in run_input)
            and importlib.util.find_spec("qiskit_ionq") is not None
        ):
            gateset = gateset or GateSet.QIS
            ionq_compiler_synthesis = ionq_compiler_synthesis or False
            run_input_compat = self._apply_qiskit_ionq_conversion(
                run_input, gateset=gateset.value, ionq_compiler_synthesis=ionq_compiler_synthesis
            )
        else:
            if gateset is not None:
                warnings.warn(
                    UserWarning(
                        "GateSet argument is only applicable when qiskit-ionq "
                        "is installed, and when all run_inputs are of type "
                        "qiskit.QuantumCircuit. Ignoring..."
                    )
                )
            if ionq_compiler_synthesis is not None:
                warnings.warn(
                    UserWarning(
                        "IonQ compiler synthesis option is only applicable when "
                        "qiskit-ionq is installed, and when all run_inputs are of "
                        "type qiskit.QuantumCircuit. Ignoring..."
                    )
                )
            run_input_compat = [self.apply_runtime_profile(program) for program in run_input]

        run_input_compat = run_input_compat[0] if is_single_input else run_input_compat
        return self.submit(run_input_compat, *args, **kwargs)



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Functions and decorators for efficient caching to improve function and method performance.
Includes per-instance LRU caching, TTL expiration, and customizable caching for specific needs.

"""
import functools
import hashlib
import json
import os
import time
from contextlib import contextmanager
from typing import Any, Callable, Generator, Optional, TypeVar, overload

TFunc = TypeVar("TFunc", bound=Callable)


_CACHE_REGISTRY = []


def _generate_cache_key(instance: Any, func_name: str, args: tuple, kwargs: dict) -> str:
    """Generate a cache key based on the class name, function name, args, and kwargs."""
    key_data = {
        "class_name": instance.__class__.__name__,
        "func_name": func_name,
        "args": args,
        "kwargs": kwargs,
    }
    key_str = json.dumps(key_data, sort_keys=True)
    return hashlib.sha256(key_str.encode()).hexdigest()


def _cached_method_wrapper(
    ttl: int = 120, maxsize: Optional[int] = 128, typed: bool = False
) -> Callable:
    """A decorator to cache the results of methods with optional TTL, maxsize, and typed options."""

    def decorator(func: Callable[..., Any]) -> Callable[..., Any]:

        @functools.lru_cache(maxsize=maxsize, typed=typed)
        def cached_func(self, *args, **kwargs):
            return func(self, *args, **kwargs)

        @functools.wraps(func)
        def wrapper(self, *args, **kwargs):
            if os.getenv("DISABLE_CACHE") == "1":
                return func(self, *args, **kwargs)
            key = _generate_cache_key(self, func.__name__, args, kwargs)

            if not getattr(self, "__cache_disabled", False) and key in cached_func.cache:
                cached_result, timestamp = cached_func.cache[key]
                if (time.time() - timestamp) < ttl:
                    return cached_result

            # The _QBRAID_TEST_CACHE_CALLS environment variable is used internally to enable
            # testing of function call counts with unittest. Since cached_func is an lru_cache
            # object, calls to it don't affect the Mock object's call count. To test call counts
            # accurately, we need to make a duplicate call the original function.

            if os.getenv("_QBRAID_TEST_CACHE_CALLS") == "1":
                _ = func(self, *args, **kwargs)
            result = cached_func(self, *args, **kwargs)
            cached_func.cache[key] = (result, time.time())
            return result

        cached_func.cache = {}
        wrapper.cache_clear = cached_func.cache_clear
        wrapper.cache_info = cached_func.cache_info
        wrapper.cache = cached_func.cache

        _CACHE_REGISTRY.append(wrapper.cache_clear)

        return wrapper

    return decorator


# Function signatures with bounded TypeVar and overload pattern inspired by The Cirq Developers
# https://github.com/quantumlib/Cirq/blob/v1.4.1/cirq-core/cirq/_compat.py


@overload
def cached_method(__func: TFunc) -> TFunc: ...


@overload
def cached_method(*, maxsize: int = 128, ttl: int = 120) -> Callable[[TFunc], TFunc]: ...


def cached_method(
    func: Optional[TFunc] = None, *, maxsize: int = 128, ttl: int = 120
) -> Callable[[TFunc], TFunc]:
    """
    AD decorator that applies default caching behavior when used without arguments,
    or allows customization (e.g., maxsize, ttl) when used with arguments.

    Example usage:

    .. code-block:: python

        @cached_method
        def some_method(self, param):
            pass

        @cached_method(maxsize=200, ttl=300)
        def some_other_method(self, param):
            pass
    """

    def decorator(inner_func: TFunc) -> TFunc:
        return _cached_method_wrapper(ttl=ttl, maxsize=maxsize)(inner_func)

    return decorator if func is None else decorator(func)


@contextmanager
def cache_disabled(instance) -> Generator[None, None, None]:
    """
    Context manager to temporarily disable caching for a specific instance.

    Args:
        instance: The class instance for which the cache should be disabled.

    Example usage:

    .. code-block:: python

        with cache_disabled(instance):
            instance.method_call()
    """
    original_value = getattr(instance, "__cache_disabled", False)
    instance.__cache_disabled = True
    try:
        yield
    finally:
        instance.__cache_disabled = original_value


def clear_cache():
    """
    Clear all least-recently-used (LRU) caches that have been registered with the
    :py:func:`qbraid._caching.cached_method` decorator.

    Use this function to completely reset the cache state for all decorated methods, which
    can be useful in testing environments or when you need to free up memory by discarding
    cached results.
    """
    for cache_clear in _CACHE_REGISTRY:
        cache_clear()



# Copyright (C) 2025 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Unit tests for the chain_calls function in qbraid.transpiler.converter.

"""
from cirq import Circuit
from openqasm3.ast import Program

from qbraid.transpiler.converter import chain_calls, translate

from .cirq_utils import _equal


def test_chain_calls_addition():
    """Test that chain_calls can add a sequence of numbers."""

    def add(x, y):
        return x + y

    result = chain_calls(add, 0, 1, 2, 3, 4)
    assert result == 10


def test_chain_calls_multiplication():
    """Test that chain_calls can multiply a sequence of numbers."""

    def multiply(x, y):
        return x * y

    result = chain_calls(multiply, 1, 2, 3, 4)
    assert result == 24


def test_chain_calls_with_kwargs():
    """Test that chain_calls can handle keyword arguments."""

    def power(x, y, exponent=1):
        return x + (y**exponent)

    result = chain_calls(power, 0, 2, 3, exponent=2)
    assert result == 13


def test_chain_calls_empty_args():
    """Test that chain_calls can handle an empty sequence of arguments."""

    def add(x, y):
        return x + y

    result = chain_calls(add, 5)
    assert result == 5


def test_chain_calls_non_numeric():
    """Test that chain_calls can concatenate a sequence of strings."""

    def concatenate(x, y):
        return x + y

    result = chain_calls(concatenate, "Hello", " ", "World", "!")
    assert result == "Hello World!"


def test_translate():
    """Test that translate can chain multiple transpile calls on a program."""
    qasm = """
    OPENQASM 2.0;
    include "qelib1.inc";
    qreg q[2];
    creg c[2];
    h q[0];
    cx q[0], q[1];
    measure q -> c;
    """

    program = translate(qasm, "openqasm3")
    circuit_one = translate(program, "qasm2", "cirq")
    circuit_two = translate(program, "qasm3", "braket", "cirq")

    assert isinstance(program, Program)
    assert all(isinstance(circuit, Circuit) for circuit in (circuit_one, circuit_two))
    assert _equal(circuit_one, circuit_two)



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Unit tests for conversions between Cirq circuits and pytket circuits.

"""
import importlib.util

import numpy as np
import pytest
from cirq import Circuit, LineQubit, ops, testing

from qbraid.interface import circuits_allclose
from qbraid.transpiler import ProgramConversionError, transpile

pytket_not_installed = importlib.util.find_spec("pytket") is None
pytestmark = pytest.mark.skipif(pytket_not_installed, reason="pytket not installed")


def test_bell_state_to_pytket():
    """Tests Circuit --> pytket.circuit.Circuit --> Circuit
    with a Bell state circuit.
    """
    qreg = LineQubit.range(2)
    cirq_circuit = Circuit([ops.H.on(qreg[0]), ops.CNOT.on(qreg[0], qreg[1])])
    pytket_circuit = transpile(cirq_circuit, "pytket")
    assert circuits_allclose(pytket_circuit, cirq_circuit, strict_gphase=True)


@pytest.mark.parametrize("num_qubits", [1, 2, 3, 4, 5])
def test_random_circuit_to_pytket(num_qubits):
    """Test converting random Cirq circuits to pytket."""
    for _ in range(10):
        cirq_circuit = testing.random_circuit(
            qubits=num_qubits,
            n_moments=np.random.randint(1, 6),
            op_density=1,
            random_state=np.random.randint(1, 10),
        )
        pytket_circuit = transpile(cirq_circuit, "pytket")
        assert circuits_allclose(pytket_circuit, cirq_circuit, strict_gphase=True)


def test_raise_error():
    """Test raising error for unsupported gates."""
    with pytest.raises(ProgramConversionError):
        probs = np.random.uniform(low=0, high=0.5)
        cirq_circuit = Circuit(ops.PhaseDampingChannel(probs).on(*LineQubit.range(1)))
        transpile(cirq_circuit, "pytket", require_native=True)



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Unit tests for converting Bloqade programs to Amazon Braket AHS

"""
from __future__ import annotations

import importlib.util
from typing import TYPE_CHECKING

import numpy as np
import pytest

# Skip all tests if bloqade not installed
bloqade_found = importlib.util.find_spec("bloqade") is not None
if bloqade_found:
    from bloqade.analog import var  # type: ignore
    from bloqade.analog.atom_arrangement import Square  # type: ignore

if TYPE_CHECKING:
    from bloqade.analog.builder.assign import BatchAssign  # type: ignore

pytestmark = pytest.mark.skipif(not bloqade_found, reason="bloqade not installed")

from qbraid.programs.ahs.braket_ahs import BraketAHS  # noqa: F821
from qbraid.transpiler.conversions.braket_ahs import bloqade_to_braket_ahs  # noqa: F821


@pytest.fixture
def bloqade_program() -> BatchAssign:
    """Create a Bloqade program batch."""

    adiabatic_durations = [0.4, 3.2, 0.4]

    max_detuning = var("max_detuning")
    adiabatic_program = (
        Square(3, lattice_spacing="lattice_spacing")
        .rydberg.rabi.amplitude.uniform.piecewise_linear(
            durations=adiabatic_durations, values=[0.0, "max_rabi", "max_rabi", 0.0]
        )
        .detuning.uniform.piecewise_linear(
            durations=adiabatic_durations,
            values=[
                -max_detuning,  # scalar variables support direct arithmetic operations
                -max_detuning,
                max_detuning,
                max_detuning,
            ],
        )
        .assign(max_rabi=15.8, max_detuning=16.33)
        .batch_assign(lattice_spacing=np.arange(4.0, 7.0, 0.5))
    )

    return adiabatic_program


@pytest.fixture
def ahs_program_batch0() -> dict:
    return {
        "register": {
            "sites": [
                ["0.000000", "0.000000"],
                ["0.000000", "0.0000040"],
                ["0.000000", "0.0000080"],
                ["0.0000040", "0.000000"],
                ["0.0000040", "0.0000040"],
                ["0.0000040", "0.0000080"],
                ["0.0000080", "0.000000"],
                ["0.0000080", "0.0000040"],
                ["0.0000080", "0.0000080"],
            ],
            "filling": [1, 1, 1, 1, 1, 1, 1, 1, 1],
        },
        "hamiltonian": {
            "drivingFields": [
                {
                    "amplitude": {
                        "time_series": {
                            "values": ["0E+5", "1.58E+7", "1.58E+7", "0E+5"],
                            "times": ["0.000000", "4E-7", "0.0000036", "0.0000040"],
                        },
                        "pattern": "uniform",
                    },
                    "phase": {
                        "time_series": {"values": ["0", "0"], "times": ["0.000000", "0.0000040"]},
                        "pattern": "uniform",
                    },
                    "detuning": {
                        "time_series": {
                            "values": ["-1.633E+7", "-1.633E+7", "1.633E+7", "1.633E+7"],
                            "times": ["0.000000", "4E-7", "0.0000036", "0.0000040"],
                        },
                        "pattern": "uniform",
                    },
                }
            ],
            "localDetuning": [],
        },
    }


def test_convert_bloqade_to_braket_ahs(bloqade_program, ahs_program_batch0):
    """Test converting Bloqade program to Amazon Braket AHS."""
    braket_ahs_program_batch = bloqade_to_braket_ahs(bloqade_program)
    braket_ahs_program = braket_ahs_program_batch[0]
    wrapped_ahs_program = BraketAHS(braket_ahs_program)
    assert wrapped_ahs_program.to_dict() == ahs_program_batch0



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Unit tests for qbraid.programs.qasm.OpenQasm3Program

"""
import numpy as np
import pytest
from qiskit.qasm3 import dumps, loads

from qbraid.interface.random.qasm3_random import qasm3_random
from qbraid.interface.random.qiskit_random import qiskit_random
from qbraid.programs.exceptions import ProgramTypeError
from qbraid.programs.gate_model.qasm3 import OpenQasm3Program
from qbraid.programs.registry import unregister_program_type

from ..fixtures.qasm3.circuits import qasm3_bell, qasm3_shared15


def test_qasm_qubits():
    """Test getting QASM qubits"""
    program1 = OpenQasm3Program(qasm3_bell())
    program1.validate()
    assert program1.qubits == {"q": 2}

    program2 = OpenQasm3Program(qasm3_shared15())
    program2.validate()
    assert program2.qubits == {"q": 4}


def test_qasm_clbits():
    """Test getting QASM clbits"""
    program1 = OpenQasm3Program("OPENQASM 3.0; bit[2] c;")
    program1.validate()
    assert program1.clbits == {"c": 2}


def test_qasm3_num_qubits():
    """Test calculating number of qubits in qasm3 circuit"""
    num_qubits = np.random.randint(2, 10)
    depth = np.random.randint(1, 4)
    qiskit_circuit = qiskit_random(num_qubits=num_qubits, depth=depth)
    qasm3_str = dumps(qiskit_circuit)
    assert OpenQasm3Program(qasm3_str).num_qubits == num_qubits


def test_qasm3_depth_alternate_qubit_syntax():
    """Test calculating qasm depth of qasm3 circuit"""
    qasm3_str = """OPENQASM 3.0;
bit[1] __bits__;
qubit[1] __qubits__;
h __qubits__[0];
__bits__[0] = measure __qubits__[0];"""
    assert OpenQasm3Program(qasm3_str).depth == 2


def _check_output(output, expected):
    actual_circuit = loads(output)
    expected_circuit = loads(expected)
    assert actual_circuit == expected_circuit


@pytest.mark.parametrize(
    "num_qubits, depth, max_operands, seed, measure",
    [
        (
            None,
            None,
            None,
            None,
            False,
        ),  # Test case 1: Generate random circuit with default parameters
        (3, 3, 3, 42, False),  # Test case 2: Generate random circuit with specified parameters
        (None, None, None, None, True),  # Test case 3: Generate random circuit with measurement
    ],
)
def test_qasm3_random(num_qubits, depth, max_operands, seed, measure):
    """Test random circuit generation using _qasm_random"""

    circuit = qasm3_random(
        num_qubits=num_qubits, depth=depth, max_operands=max_operands, seed=seed, measure=measure
    )
    assert OpenQasm3Program(circuit).num_qubits >= 1
    if num_qubits is not None:
        assert OpenQasm3Program(circuit).num_qubits == num_qubits


def test_qasm3_random_with_known_seed():
    """Test generating random OpenQASM 3 circuit from known seed"""
    circuit = qasm3_random(num_qubits=3, depth=3, max_operands=3, seed=42, measure=True)
    assert OpenQasm3Program(circuit).num_qubits == 3

    out__expected = """
// Random Circuit generated by qBraid
OPENQASM 3.0;
include "stdgates.inc";
/*
    seed = 42
    num_qubits = 3
    depth = 3
    max_operands = 3
*/
qubit[3] q;
bit[3] c;
y q[0];
crx(5.3947298351621535) q[1],q[2];
cz q[0],q[2];
t q[1];
cp(0.8049616944763924) q[1],q[2];
u1(2.829858307545725) q[0];
c[0] = measure q[0];
c[1] = measure q[1];
c[2] = measure q[2];

"""
    _check_output(circuit, out__expected)


def test_qasm3_num_qubits_alternate_synatx():
    """Test calculating num qubits for qasm3 syntax edge-case"""
    qasm3_str = """
OPENQASM 3;
include "stdgates.inc";
qubit _qubit0;
qubit _qubit1;
h _qubit0;
cx _qubit0, _qubit1;
"""
    qprogram = OpenQasm3Program(qasm3_str)
    assert qprogram.num_qubits == 2


def test_qasm3_num_clbits():
    """Test calculating num classical bits for qasm3"""
    qasm3_str = """
OPENQASM 3;
include "stdgates.inc";
qubit[2] q;
bit[2] c;
h q[0];
cx q[0], q[1];
measure q[0] -> c[0];
measure q[1] -> c[1];
"""
    qprogram = OpenQasm3Program(qasm3_str)
    assert qprogram.num_clbits == 2


def test_populate_idle_qubits_qasm3():
    """Test conversion of qasm3 to contiguous qasm3"""
    qasm_test = """OPENQASM 3.0;
    gate custom q1, q2, q3{
        x q1;
        y q2;
        z q3;
    }
    qubit[2] q1;
    qubit[2] q2;
    qubit[3] q3;
    qubit[1] q4;
    x q1[0];
    y q2[0];
    z q3;
    """

    qprogram = OpenQasm3Program(qasm_test)
    qprogram.populate_idle_qubits()
    actual_qasm = qprogram.program

    id_ops = ["id q1[1];", "id q2[1];", "id q4[0];"]
    for id_op in id_ops:
        assert id_op in actual_qasm


def test_remove_idle_qubits_qasm3():
    """Test that remove_idle_qubits for qasm3 string"""
    qasm3_str = """
    OPENQASM 3;
    include "stdgates.inc";
    qubit[4] q;
    h q[1];
    cx q[1], q[3];
    """
    qprogram = OpenQasm3Program(qasm3_str)
    qprogram.remove_idle_qubits()
    assert qprogram.num_qubits == 2


def test_reverse_qubit_order():
    """Test the reverse qubit ordering function"""
    qasm_str = """
    OPENQASM 3.0;
    include "stdgates.inc";
    qubit[2] q;
    qubit[4] q2;
    qubit[1] q3;
    cx q[0], q[1];
    cx q2[0], q2[1];
    x q2[3];
    cx q2[0], q2[2];
    x q3[0];
    """

    program = OpenQasm3Program(qasm_str)
    program.reverse_qubit_order()
    reverse_qasm = program.program

    expected_qasm = """OPENQASM 3.0;
include "stdgates.inc";
qubit[2] q;
qubit[4] q2;
qubit[1] q3;
cx q[1], q[0];
cx q2[3], q2[2];
x q2[0];
cx q2[3], q2[1];
x q3[0];
"""
    assert reverse_qasm == expected_qasm


def test_raise_program_type_error():
    """Test that initializing OpenQasm3Program with an invalid type raises ProgramTypeError."""
    try:
        with pytest.raises(ProgramTypeError):
            OpenQasm3Program(42)
    finally:
        unregister_program_type("int")



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module defining qBraid runtime base schema,
with a shared header for semantic versioning.

"""
from __future__ import annotations

from decimal import Decimal
from typing import ClassVar, Union

import qbraid_core.decimal
from pydantic import BaseModel, Field, GetCoreSchemaHandler, computed_field, model_validator
from pydantic.json_schema import JsonSchemaValue
from pydantic_core import core_schema
from typing_extensions import Annotated, Self


class QbraidSchemaHeader(BaseModel):
    """Represents the header for a qBraid schema.

    Attributes:
        name (str): Name of the schema, typically the package and module.
        version (float): Version number of the schema.
    """

    name: Annotated[str, Field(strict=True, min_length=1)]
    version: Annotated[float, Field(strict=True, gt=0)]


class QbraidSchemaBase(BaseModel):
    """Base class for qBraid schemas that require a valid version in subclasses.

    Attributes:
        header (QbraidSchemaHeader): The schema header.
    """

    VERSION: ClassVar[float]  # Must be defined in subclasses

    @computed_field(alias="schemaHeader")
    @property
    def header(self) -> QbraidSchemaHeader:
        """Computes the schema header based on the module name and class version.

        Returns:
            QbraidSchemaHeader: The computed schema header.

        Raises:
            ValueError: If 'VERSION' is not defined or invalid.
        """
        if (
            not hasattr(self.__class__, "VERSION")
            or not isinstance(self.__class__.VERSION, (int, float))
            or self.__class__.VERSION <= 0
        ):
            raise ValueError(
                f"{self.__class__.__name__} must define a valid semantic version for 'VERSION'."
            )

        module_name = self.__class__.__module__
        return QbraidSchemaHeader(name=module_name, version=float(self.__class__.VERSION))

    @model_validator(mode="after")
    def validate_header(self) -> Self:
        """Validates that the header is correctly set up during instantiation."""
        try:
            self.header
        except ValueError as err:
            raise err
        return self


class Credits(qbraid_core.decimal.Credits):
    """Represents a value in Credits, subclassing decimal.Decimal.

    Args:
        value (str or int or float): The initial value for the Credits.
    """

    @classmethod
    def __get_pydantic_core_schema__(  # pylint: disable-next=unused-argument
        cls, source_type: type, handler: GetCoreSchemaHandler
    ) -> core_schema.CoreSchema:
        """Provide Pydantic with a schema for the Credits type."""
        number_schema = core_schema.union_schema(
            [
                core_schema.int_schema(),
                core_schema.float_schema(),
                handler(Decimal),
            ],
            custom_error_type="credits_type",
            custom_error_message="Input should be a number (int, float, or Decimal)",
        )

        # pylint: disable-next=unused-argument
        def to_credits(value: Union[int, float, Decimal], info) -> "Credits":
            return cls(value)

        return core_schema.with_info_after_validator_function(to_credits, number_schema)

    @classmethod
    def __get_pydantic_json_schema__(
        cls, core_schema: core_schema.CoreSchema, handler
    ) -> JsonSchemaValue:
        """Provide a JSON schema for the Credits type."""
        json_schema = handler(core_schema)
        json_schema.update(
            title="Credits",
            description="A monetary amount where 1 Credit = $0.01 USD.",
            examples=[10, 0.05, 1.5],
            type="number",
        )
        return json_schema


class USD(qbraid_core.decimal.USD):
    """Represents a value in U.S. Dollars, subclassing decimal.Decimal.

    Args:
        value (str or int or float): The initial value for the USD.
    """

    @classmethod
    def __get_pydantic_core_schema__(  # pylint: disable-next=unused-argument
        cls, source_type: type, handler: GetCoreSchemaHandler
    ) -> core_schema.CoreSchema:
        """Provide Pydantic with a schema for the USD type."""
        number_schema = core_schema.union_schema(
            [
                core_schema.int_schema(),
                core_schema.float_schema(),
                handler(Decimal),
            ],
            custom_error_type="usd_type",
            custom_error_message="Input should be a number (int, float, or Decimal)",
        )

        # pylint: disable-next=unused-argument
        def to_usd(value: Union[int, float, Decimal], info) -> "USD":
            return cls(value)

        return core_schema.with_info_after_validator_function(to_usd, number_schema)

    @classmethod
    def __get_pydantic_json_schema__(
        cls, core_schema: core_schema.CoreSchema, handler
    ) -> JsonSchemaValue:
        """Provide a JSON schema for the USD type."""
        json_schema = handler(core_schema)
        json_schema.update(
            title="USD",
            description="A monetary amount representing U.S. Dollars.",
            examples=[10, 0.05, 1.5],
            type="number",
        )
        return json_schema



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module defining BraketCircuit Class

"""
from __future__ import annotations

from typing import TYPE_CHECKING

from braket.circuits import Circuit, Instruction, Qubit
from braket.circuits.measure import Measure

from qbraid.programs.exceptions import ProgramTypeError

from ._model import GateModelProgram

if TYPE_CHECKING:
    import braket.circuits
    import numpy as np


class BraketCircuit(GateModelProgram):
    """Wrapper class for ``braket.circuits.Circuit`` objects."""

    def __init__(self, program: braket.circuits.Circuit):
        super().__init__(program)
        if not isinstance(program, Circuit):
            raise ProgramTypeError(
                message=f"Expected 'braket.circuits.Circuit' object, got '{type(program)}'."
            )

    @property
    def qubits(self) -> list[Qubit]:
        """Return the qubits acted upon by the operations in this circuit"""
        return list(self.program.qubits)

    @property
    def num_qubits(self) -> int:
        """Return the number of qubits in the circuit."""
        return self.program.qubit_count

    @property
    def num_clbits(self) -> int:
        """Return the number of classical bits in the circuit."""
        return 0

    @property
    def depth(self) -> int:
        """Return the circuit depth (i.e., length of critical path)."""
        return self.program.depth

    def _unitary(self) -> np.ndarray:
        """Calculate unitary of circuit."""
        return self.program.to_unitary()

    def populate_idle_qubits(self) -> None:
        """Checks whether the circuit uses contiguous qubits/indices,
        and if not, adds identity gates to vacant registers as needed."""
        max_qubit = 0
        occupied_qubits = []
        circuit = self.program.copy()
        for qubit in circuit.qubits:
            index = int(qubit)
            occupied_qubits.append(index)
            max_qubit = max(max_qubit, index)
        qubit_count = max_qubit + 1
        if qubit_count > circuit.qubit_count:
            all_qubits = list(range(0, qubit_count))
            vacant_qubits = list(set(all_qubits) - set(occupied_qubits))
            for index in vacant_qubits:
                circuit.i(index)
        self._program = circuit

    def remove_idle_qubits(self) -> None:
        """Checks whether the circuit uses contiguous qubits/indices,
        and if not, reduces dimension accordingly."""
        qubit_map = {}
        circuit = self.program.copy()
        circuit_qubits = list(circuit.qubits)
        circuit_qubits.sort()
        for index, qubit in enumerate(circuit_qubits):
            qubit_map[int(qubit)] = index
        contig_circuit = Circuit()
        for instr in circuit.instructions:
            contig_target = [qubit_map[int(qubit)] for qubit in list(instr.target)]
            contig_control = [qubit_map[int(qubit)] for qubit in list(instr.control)]
            contig_instr = Instruction(
                instr.operator,
                target=contig_target,
                control=contig_control,
                control_state=instr.control_state,
                power=instr.power,
            )
            contig_circuit.add_instruction(contig_instr)
        self._program = contig_circuit

    def reverse_qubit_order(self) -> None:
        """Reverse the order of the qubits in the circuit."""
        qubit_map = {}
        circuit = self.program.copy()
        circuit_qubits = list(circuit.qubits)
        circuit_qubits.sort()
        circuit_qubits = list(reversed(circuit_qubits))
        for index, qubit in enumerate(circuit_qubits):
            qubit_map[int(qubit)] = index
        contig_circuit = Circuit()
        for instr in circuit.instructions:
            contig_target = [qubit_map[int(qubit)] for qubit in list(instr.target)]
            contig_control = [qubit_map[int(qubit)] for qubit in list(instr.control)]
            contig_instr = Instruction(
                instr.operator,
                target=contig_target,
                control=contig_control,
                control_state=instr.control_state,
                power=instr.power,
            )
            contig_circuit.add_instruction(contig_instr)
        self._program = contig_circuit

    def pad_measurements(self) -> None:
        """
        Pad the circuit with measurements on all qubits and track partial measurements.

        It adds partial measurement support to device that requires measuring all qubits
        (e.g., IonQ devices).

        This method identifies qubits that already have measurement instructions (partial
        measurements) and adds measurement instructions to all remaining qubits. It stores
        the list of originally measured qubits as an attribute for later use in result
        processing.

        The method modifies the circuit in-place and sets the `partial_measurement_qubits`
        attribute on the program object to track which qubits were originally measured
        before padding. The padding only occur when there is partial measurement in a
        circuit. If there is no measurement int the circuit, the backend assumes measuring
        all qubits and no padding is applied.
        """
        # Track qubits that already have measurement instructions
        partial_measurement_qubits: list[int] = []
        for instruction in self._program.instructions:
            if isinstance(instruction.operator, Measure):
                partial_measurement_qubits.append(int(instruction.target[0]))

        # Only apply padding when there is partial measurement
        if len(partial_measurement_qubits) == 0:
            return

        # Add measurements to all qubits that don't already have them
        for qubit in self._program.qubits:
            if qubit not in partial_measurement_qubits:
                self._program.measure(qubit)

        # Store the original partial measurement qubits for result processing
        self._program.partial_measurement_qubits = partial_measurement_qubits

    def transform(self, device) -> None:
        """Transform program to according to device target profile."""
        if device.simulator:
            self.remove_idle_qubits()

        # For IonQ and Amazon Braket simulators, pad measurements to support partial measurement
        if device._provider_name in ["IonQ", "Amazon Braket"]:
            self.pad_measurements()

    def serialize(self) -> dict[str, str]:
        """Return the program in a format suitable for submission to the qBraid API."""
        # pylint: disable=import-outside-toplevel
        from qbraid.programs.gate_model.qasm3 import OpenQasm3Program
        from qbraid.transpiler.conversions.braket import braket_to_qasm3

        # pylint: enable=import-outside-toplevel

        qasm = braket_to_qasm3(self.program)
        program = OpenQasm3Program(qasm)

        return program.serialize()



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module containing function to convert from Cirq's circuit
representation to pyQuil's circuit representation (Quil programs).

"""
from __future__ import annotations

from typing import TYPE_CHECKING

from cirq import LineQubit, QubitOrder
from qbraid_core._import import LazyLoader

from qbraid.transpiler.annotations import weight
from qbraid.transpiler.exceptions import ProgramConversionError

try:
    from .cirq_quil_output import QuilOutput
except ImportError:  # pragma: no cover
    QuilOutput = None

pyquil = LazyLoader("pyquil", globals(), "pyquil")

if TYPE_CHECKING:
    import cirq.circuits
    from pyquil import Program


@weight(0.74)
def cirq_to_pyquil(circuit: cirq.circuits.Circuit) -> Program:
    """Returns a pyQuil Program equivalent to the input Cirq circuit.

    Args:
        circuit: Cirq circuit to convert to a pyQuil Program.

    Returns:
        pyquil.Program object equivalent to the input Cirq circuit.
    """
    input_qubits = circuit.all_qubits()
    max_qubit = max(input_qubits)
    # if we are using LineQubits, keep the qubit labeling the same
    if isinstance(max_qubit, LineQubit):
        qubit_range = max_qubit.x + 1
        qubit_order = LineQubit.range(qubit_range)
    # otherwise, use the default ordering (starting from zero)
    else:
        qubit_order = QubitOrder.DEFAULT
    qubits = QubitOrder.as_qubit_order(qubit_order).order_for(input_qubits)
    operations = circuit.all_operations()
    try:
        quil_str = str(QuilOutput(operations, qubits))
        return pyquil.Program(quil_str)
    except ValueError as err:
        raise ProgramConversionError(
            f"Cirq qasm converter doesn't yet support {err.args[0][32:]}."
        ) from err



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module defining qBraid runtime experiment schemas.

"""
from __future__ import annotations

from collections import Counter
from typing import TYPE_CHECKING, Any, Optional, Union

from pydantic import BaseModel, ConfigDict, Field, field_validator, model_validator

from qbraid.programs import load_program
from qbraid.programs.typer import Qasm2String, Qasm3String

if TYPE_CHECKING:
    from typing_extensions import Self


class ExperimentMetadata(BaseModel):
    """Base class for experiment metadata.

    This class serves as a base for more specific experiment metadata classes.
    """

    model_config = ConfigDict(extra="allow")


class GateModelExperimentMetadata(BaseModel):
    """Metadata specific to gate-model experiments, i.e. experiments defined
    using an intermediate representation (IR) that is compatible with OpenQASM.

    Attributes:
        measurement_counts (Counter): Counter for measurement outcomes.
        measurements (list, optional): Optional list of measurement results.
        qasm (str, optional): OpenQASM string representation of the quantum circuit.
        num_qubits (int, optional): Number of qubits used in the circuit.
        gate_depth (int, optional): Depth of the quantum circuit (i.e. length of critical path)
    """

    measurement_counts: Optional[Counter] = Field(None, alias="measurementCounts")
    measurements: Optional[list] = None
    qasm: Optional[str] = Field(None, alias="openQasm")
    num_qubits: Optional[int] = Field(None, alias="circuitNumQubits")
    gate_depth: Optional[int] = Field(None, alias="circuitDepth")

    @field_validator("measurement_counts")
    @classmethod
    def validate_counts(cls, value):
        """Validates and ensures that the measurement counts are a Counter object.

        Args:
            value: The measurement counts.

        Returns:
            Counter: A validated counter object.
        """
        return Counter(value)

    @field_validator("qasm")
    @classmethod
    def validate_qasm(cls, value):
        """Validates that the provided QASM is valid.

        Args:
            value: The QASM string.

        Returns:
            The validated QASM string.

        Raises:
            ValueError: If the QASM string is not valid.
        """
        if value is None:
            return value
        if not isinstance(value, (Qasm2String, Qasm3String)):
            raise ValueError("openQasm must be a valid OpenQASM string.")
        return value

    @model_validator(mode="after")
    def set_qubits_and_depth(self) -> Self:
        """Sets the number of qubits and gate depth if not already provided.

        Returns:
            Self: The updated instance with qubits and depth set.
        """
        if self.qasm and (self.num_qubits is None or self.gate_depth is None):
            program = load_program(self.qasm)
            self.num_qubits = self.num_qubits or program.num_qubits
            self.gate_depth = self.gate_depth or program.depth  # type: ignore

        return self


class QbraidQirSimulationMetadata(GateModelExperimentMetadata):
    """Metadata specific to jobs submitted to the qBraid QIR simulator.

    Attributes:
        backend_version (str, optional): The version of the simulator backend.
        seed (int, optional): The seed used for the simulation.

    """

    backend_version: Optional[str] = Field(None, alias="runnerVersion")
    seed: Optional[int] = Field(None, alias="runnerSeed")


class QuEraQasmSimulationMetadata(GateModelExperimentMetadata):
    """Metadata specific to jobs submitted to the QuEra QASM simulator.

    Attributes:
        backend (str, optional): The name of the backend used for the simulation.
        quera_simulation_result (dict[str, Any], optional): The result of the simulation
            in a dictionary format. This includes data describing the state of the
            QPU atoms used in the simulation and logs generated during the simulation,
            among other things.

    """

    backend: Optional[str] = None
    quera_simulation_result: Optional[dict[str, Any]] = None


class Equal1SimulationMetadata(GateModelExperimentMetadata):
    """Metadata specific to jobs submitted to the Equal1 simulator.

    Attributes:
        compiled_output (str, optional): The compiled output of the job.
        execution_mode (str, optional): The execution mode of the job.
        device_name (str, optional): The name of the device used for the job.
        simulation_platform (str, optional): The platform used for the simulation.
        simulation_mode (str, optional): The mode used for the simulation.
        execution_options (dict[str, Any], optional): The execution options used for the job.
    """

    compiled_output: Optional[str] = Field(None, alias="compiledOutput")
    execution_mode: Optional[str] = Field(None, alias="executionMode")
    device_name: Optional[str] = Field(None, alias="deviceName")
    simulation_platform: Optional[str] = Field(None, alias="simulationPlatform")
    simulation_mode: Optional[str] = Field(None, alias="simulationMode")
    execution_options: Optional[dict[str, Any]] = Field(None, alias="executionOptions")


class AhsExperimentMetadata(BaseModel):
    """Metadata specific to Analog Hamiltonian Simulation (AHS) experiments.

    Attributes:
        measurement_counts (Counter): Counter for measurement outcomes
        measurements (list, optional): Optional list of measurement results
        num_atoms (int, optional): Number of atoms (sites) used to build lattice structure
        sites (list[tuple[float, float]], optional): Vector positions of atoms in meters
        filling (list[int], optional): List of ints {0,1} indicating the filling status at each site
    """

    measurement_counts: Optional[Counter] = Field(None, alias="measurementCounts")
    measurements: Optional[list[dict[str, Union[bool, Optional[list[int]]]]]] = None
    num_atoms: Optional[int] = Field(None, alias="numAtoms")
    sites: Optional[list[tuple[float, float]]] = None
    filling: Optional[list[int]] = None

    @field_validator("filling")
    @classmethod
    def validate_filling(cls, value):
        """Validates and ensures that the fillings are integers 0 or 1.

        Args:
            value: The filling status at each site.

        Returns:
            list[int]: A validated list of integers.
        """
        if value is None:
            return value

        def validate_item(item):
            int_item = int(item)
            if int_item not in {0, 1}:
                raise ValueError(f"Invalid filling value: {item}. Must be 0 or 1.")
            return int_item

        try:
            return [validate_item(filling) for filling in value]
        except ValueError as err:
            raise ValueError(
                "Invalid filling value. Must be a list of integers, each either 0 or 1."
            ) from err

    @model_validator(mode="after")
    def validate_ahs(self) -> Self:
        """Validates that the sites, filling, and num_atoms lengths match.

        Returns:
            Self: The updated instance with validated lengths.
        """
        lengths = {
            "sites": len(self.sites) if self.sites else None,
            "filling": len(self.filling) if self.filling else None,
            "num_atoms": self.num_atoms,
        }

        filtered_lengths = {k: v for k, v in lengths.items() if v is not None}

        if len(set(filtered_lengths.values())) > 1:
            mismatched = ", ".join(f"{k}: {v}" for k, v in filtered_lengths.items())
            raise ValueError(
                "The lengths of 'sites', 'filling', and value of 'num_atoms' must be consistent. "
                f"Detected mismatched values: {mismatched}"
            )

        if self.num_atoms is None and filtered_lengths:
            self.num_atoms = next(iter(filtered_lengths.values()))

        return self


class AnnealingExperimentMetadata(BaseModel):
    """Metadata specific to annealing experiments."""

    solutions: Optional[list[dict[str, Any]]] = None
    num_solutions: Optional[int] = Field(None, alias="solutionCount")
    energies: Optional[list[float]] = None
    num_variables: Optional[int] = Field(None, alias="numVariables")


class QuboSolveParams(BaseModel):
    """Parameters for solving a QUBO problem using NEC VA algorithm.

    Attributes:
        offset (float): Offset for the normalized weight information stored in the QUBO.
        num_reads (Optional[int]): VA sampling rate. Must be between 1 and 20. Default is 1.
        num_results (Optional[int]): Number of VA annealing results. Returns only the optimal
            solution when 1 or None is specified. Default is 1.
        num_sweeps (Optional[int]): Number of VA annealing sweeps. Must be between 1 and 100000.
            Default is 500.
        beta_range (Optional[tuple[float, float, int]]): VA beta value in (start, end, steps)
            format. Default is (10.0, 100.0, 200).
        beta_list (Optional[list[float]]): Beta value array for each VA sweep.
        dense (Optional[bool]): VA matrix mode. True for dense matrix mode, False for sparse
            matrix mode. Default is None.
        vector_mode (Optional[str]): Mode during VA annealing. Options are 'speed' for speed
            priority or 'accuracy' for accuracy priority. Default is 'accuracy'.
        timeout (Optional[int]): Job execution timeout in seconds. Standard range is between 1
            and 7200. Default is 1800.
        ve_num (Optional[int]): Number of VEs used in VA annealing. Must be between 1 and the
            number of VEs installed on each server.
        onehot (Optional[list[list[str]]]): VA onehot constraint.
        fixed (Optional[Union[dict[str, int], list[list[Union[str, int]]]]): VA fixed constraint.
        andzero (Optional[list[list[str]]]): VA andzero constraint.
        orone (Optional[list[list[str]]]): VA orone constraint.
        supplement (Optional[list[list[str]]]): VA supplement constraint.
        maxone (Optional[list[list[Union[int, list[str]]]]): VA maxone constraint.
        minmaxone (Optional[list[list[Union[int, list[str]]]]): VA minmaxone constraint.
        init_spin (Optional[Union[dict[str, int], list[list[Union[str, int]]]]):
            VA initial spin parameter.
        spin_list (Optional[list[str]]): VA spin list parameter.
    """

    offset: float

    num_reads: Optional[int] = 1
    num_results: Optional[int] = 1
    num_sweeps: Optional[int] = 500
    beta_range: Optional[tuple[float, float, int]] = (10.0, 100.0, 200)
    beta_list: Optional[list[float]] = None
    dense: Optional[bool] = None

    vector_mode: Optional[str] = "accuracy"
    timeout: Optional[int] = 1800
    ve_num: Optional[int] = None
    onehot: Optional[list[list[str]]] = None
    fixed: Optional[Union[dict[str, int], list[list[Union[str, int]]]]] = None
    andzero: Optional[list[list[str]]] = None
    orone: Optional[list[list[str]]] = None
    supplement: Optional[list[list[str]]] = None
    maxone: Optional[list[list[Union[int, list[str]]]]] = None
    minmaxone: Optional[list[list[Union[int, list[str]]]]] = None
    init_spin: Optional[Union[dict[str, int], list[list[Union[str, int]]]]] = None
    spin_list: Optional[list[str]] = None

    @field_validator("offset")
    @classmethod
    def validate_offset(cls, value):
        """Validate the offset value."""
        if not -3.402823e38 <= value <= 3.402823e38:
            raise ValueError("offset must be between -3.402823e+38 and 3.402823e+38")
        return value

    @field_validator("num_reads")
    @classmethod
    def validate_num_reads(cls, value):
        """Validate the num_reads value."""
        if value is not None and not 1 <= value <= 20:
            raise ValueError("num_reads must be between 1 and 20")
        return value

    @field_validator("num_sweeps")
    @classmethod
    def validate_num_sweeps(cls, value):
        """Validate the num_sweeps value."""
        if value is not None and not 1 <= value <= 100000:
            raise ValueError("num_sweeps must be between 1 and 100000")
        return value

    @field_validator("beta_range")
    @classmethod
    def validate_beta_range(cls, value):
        """Validate the beta_range value."""
        start, end, steps = value

        min_value = 1.1754945e-38
        max_value = 3.402823e38

        if not min_value <= start <= max_value:
            raise ValueError(f"start value must be between {min_value} and {max_value}")

        if not min_value <= end <= max_value:
            raise ValueError(f"end value must be between {min_value} and {max_value}")

        if start > end:
            raise ValueError("start value must be less than or equal to end value")

        if not 1 <= steps <= 100000:
            raise ValueError("steps must be between 1 and 100000")

        return value

    @field_validator("beta_list")
    @classmethod
    def validate_beta_list(cls, value):
        """Validate the beta_list value."""
        if value is None:
            return value

        min_value = 1.1754945e-38
        max_value = 3.402823e38

        for beta in value:
            if not min_value <= beta <= max_value:
                raise ValueError(
                    f"All beta values must be between {min_value} and {max_value}. "
                    f"Found invalid value: {beta}"
                )

        return value

    @field_validator("timeout")
    @classmethod
    def validate_timeout(cls, value):
        """Validate the timeout value."""
        if value is not None and not 1 <= value <= 7200:
            raise ValueError("timeout must be between 1 and 7200 seconds")
        return value

    @field_validator("vector_mode")
    @classmethod
    def validate_vector_mode(cls, value):
        """Validate the vector_mode value."""
        if value is not None and value not in {"speed", "accuracy"}:
            raise ValueError("vector_mode must be 'speed' or 'accuracy'")
        return value

    @field_validator("ve_num")
    @classmethod
    def validate_ve_num(cls, value):
        """Validate the ve_num value."""
        if value is not None and value < 1:
            raise ValueError("ve_num must be greater than or equal to 1")
        return value



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

# pylint: disable=redefined-outer-name

"""
Unit tests for equivalence of interfacing quantum programs

"""
from __future__ import annotations

from unittest.mock import Mock

import numpy as np
import pytest
from braket.circuits import Circuit as BKCircuit
from cirq import Circuit, LineQubit, X, Y, Z

from qbraid.interface.circuit_equality import circuits_allclose
from qbraid.interface.random.random import random_circuit, random_unitary_matrix
from qbraid.programs import get_program_type_alias, load_program
from qbraid.programs.ahs import AnalogHamiltonianProgram
from qbraid.programs.exceptions import ProgramTypeError
from qbraid.programs.gate_model import GateModelProgram

from ..fixtures import packages_bell, packages_shared15


def pair_packages(packages):
    """Return a list of tuples of packages to compare"""
    return [(packages[i], packages[i + 1]) for i in range(len(packages) - 1)]


bell_pairs = pair_packages(packages_bell)
shared15_pairs = pair_packages(packages_shared15)


@pytest.mark.parametrize("two_bell_circuits", bell_pairs, indirect=True)
def test_compare_bell_circuits(two_bell_circuits):
    """Test unitary equivalance of bell circuits across packages for
    testing baseline."""
    circuit1, circuit2, _, _ = two_bell_circuits
    try:
        equal = circuits_allclose(circuit1, circuit2)
        assert equal
    except NotImplementedError:
        pytest.skip("Not implemented")


@pytest.mark.parametrize("two_shared15_circuits", shared15_pairs, indirect=True)
def test_compare_shared15_circuits(two_shared15_circuits):
    """Test unitary equivalance of shared15 circuits across packages for
    testing baseline."""
    circuit1, circuit2, package1, package2 = two_shared15_circuits
    strict_gphase = not (
        "pyquil" in {package1, package2}
        or {package1, package2} == {"qasm2", "qasm3"}
        or {package1, package2} == {"pytket", "qasm2"}
    )
    try:
        equal = circuits_allclose(circuit1, circuit2, strict_gphase=strict_gphase)
        assert equal
    except NotImplementedError:
        pytest.skip("Not implemented")


@pytest.mark.parametrize("package", ["braket", "cirq", "qiskit"])
def test_random(package):
    """Test generating random circuits"""
    program = random_circuit(package)
    assert get_program_type_alias(program) == package


def test_collapse_empty_braket_cirq():
    """Test unitary equivalance after converting to contiguous qubits"""
    # pylint: disable=no-member
    braket_circuit = BKCircuit()
    braket_circuit.x(0)
    braket_circuit.y(2)
    braket_circuit.z(4)
    # pylint: enable=no-member
    assert braket_circuit.qubit_count == 3

    cirq_circuit = Circuit()
    q0 = LineQubit(0)
    q2 = LineQubit(2)
    q4 = LineQubit(4)
    cirq_circuit.append(X(q0))
    cirq_circuit.append(Y(q2))
    cirq_circuit.append(Z(q4))
    assert len(cirq_circuit.all_qubits()) == 3

    assert circuits_allclose(braket_circuit, cirq_circuit, strict_gphase=True)

    qprogram = load_program(braket_circuit)
    qprogram.remove_idle_qubits()
    braket_compat_circuit = qprogram.program
    assert braket_compat_circuit.qubit_count == 3

    qprogram = load_program(cirq_circuit)
    qprogram.remove_idle_qubits()
    cirq_compat_circuit = qprogram.program
    assert circuits_allclose(braket_compat_circuit, cirq_compat_circuit, strict_gphase=True)

    qprogram = load_program(cirq_circuit)
    qprogram.populate_idle_qubits()
    cirq_expanded_circuit = qprogram.program
    assert len(cirq_expanded_circuit.all_qubits()) == 5


def test_random_unitary():
    """Test generating random unitary"""
    matrix = random_unitary_matrix(2)
    assert np.allclose(matrix @ matrix.conj().T, np.eye(2))


class FakeProgram(GateModelProgram):
    """Fake program for testing"""

    def __init__(self):
        super().__init__("OPENQASM 2.0;qreg q[2];h q[0];cx q[0],q[1];")

    @property
    def qubits(self):
        raise NotImplementedError

    @property
    def num_clbits(self):
        """Return the number of classical bits in the circuit."""
        raise NotImplementedError

    @property
    def depth(self) -> int:
        """Return the circuit depth (i.e., length of critical path)."""
        raise NotImplementedError

    def _unitary(self) -> np.ndarray:
        """Calculate unitary of circuit."""
        raise NotImplementedError

    def remove_idle_qubits(self):
        """Remove empty registers of circuit."""
        raise NotImplementedError

    def reverse_qubit_order(self):
        """Rerverse qubit ordering of circuit."""
        raise NotImplementedError


@pytest.fixture
def fake_program():
    """Return a FakeProgram object."""
    return FakeProgram()


@pytest.mark.parametrize(
    "matrix",
    [
        np.array([[1, 2, 3], [4, 5, 6]]),  # Non-square matrix
        np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]]),  # Square but not a power of 2
    ],
)
def test_unitary_rev_qubits_value_errors(matrix, fake_program):
    """Test that unitary_rev_qubits raises a ValueError for invalid input matrices."""
    expected_error_msg = "Input matrix must be a square matrix of size 2^N for some integer N."
    with pytest.raises(ValueError) as excinfo:
        fake_program._unitary = lambda: matrix
        fake_program.unitary_rev_qubits()
    assert expected_error_msg in str(excinfo.value)


@pytest.mark.parametrize(
    "matrix",
    [
        np.array([[1, 2], [3, 4]]),  # Not a unitary matrix, simple 2x2 case
        np.array([[0, 1], [1, 0], [0, 0]]),  # Non-square matrix which cannot be unitary
    ],
)
def test_unitary_little_endian_non_unitary_matrix_raises_value_error(matrix, fake_program):
    """Test that unitary_little_endian raises a ValueError for non-unitary matrices."""
    expected_error_msg = "Input matrix must be unitary."
    with pytest.raises(ValueError) as excinfo:
        fake_program.unitary = lambda: matrix
        fake_program.unitary_little_endian()
    assert expected_error_msg in str(excinfo.value)


def test_program_type_error():
    """Test error setting different type of same program"""
    braket_circuit = BKCircuit()
    braket_circuit.x(0)
    braket_circuit.y(2)
    braket_circuit.z(4)

    qbraid_circuit = load_program(braket_circuit)
    with pytest.raises(ProgramTypeError):
        qbraid_circuit.program = "string"


def test_ahs_program_transform():
    """Test that AHS transform method does not modify the program."""

    class MyAHS(AnalogHamiltonianProgram):
        """Fake AnalogHamiltonianProgram for testing."""

        def to_dict(self):
            """Mock to_dict method."""
            return self.program

    ahs_json = {"hamiltonian": {"terms": []}, "register": {"sites": [], "filling": []}}
    program = MyAHS(ahs_json)
    program.transform(device=Mock())
    assert program.program == ahs_json



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
This top level module contains the main qBraid public functionality.

.. currentmodule:: qbraid

Functions
----------

.. autosummary::
   :toctree: ../stubs/

    about
    clear_cache
    cache_disabled

Exceptions
-----------

.. autosummary::
   :toctree: ../stubs/

   QbraidError

"""
import importlib
from typing import TYPE_CHECKING

from ._about import about
from ._caching import cache_disabled, clear_cache
from ._version import __version__
from .exceptions import QbraidError

__all__ = [
    "about",
    "QbraidError",
    "__version__",
    "clear_cache",
    "cache_disabled",
]

_lazy = {
    "interface": [
        "circuits_allclose",
        "random_circuit",
    ],
    "passes": [],
    "programs": [
        "AnalogHamiltonianProgram",
        "GateModelProgram",
        "ProgramSpec",
        "Qasm2String",
        "Qasm3String",
        "QuantumProgram",
        "QPROGRAM",
        "QPROGRAM_NATIVE",
        "QPROGRAM_REGISTRY",
        "NATIVE_REGISTRY",
        "get_program_type_alias",
        "load_program",
        "register_program_type",
        "unregister_program_type",
        "ExperimentType",
    ],
    "runtime": [
        "AhsResultData",
        "AhsShotResult",
        "DeviceStatus",
        "GateModelResultData",
        "JobStatus",
        "NoiseModel",
        "QbraidProvider",
        "QbraidDevice",
        "QbraidJob",
        "QbraidSession",
        "QbraidClient",
        "QuantumDevice",
        "QuantumJob",
        "QuantumProvider",
        "Result",
        "ResultData",
        "RuntimeOptions",
        "Session",
        "TargetProfile",
    ],
    "transpiler": ["Conversion", "ConversionGraph", "ConversionScheme", "transpile", "translate"],
    "visualization": [],
}

if TYPE_CHECKING:
    from .interface import circuits_allclose as circuits_allclose
    from .interface import random_circuit as random_circuit
    from .programs import NATIVE_REGISTRY as NATIVE_REGISTRY
    from .programs import QPROGRAM as QPROGRAM
    from .programs import QPROGRAM_NATIVE as QPROGRAM_NATIVE
    from .programs import QPROGRAM_REGISTRY as QPROGRAM_REGISTRY
    from .programs import AnalogHamiltonianProgram as AnalogHamiltonianProgram
    from .programs import ExperimentType as ExperimentType
    from .programs import GateModelProgram as GateModelProgram
    from .programs import ProgramSpec as ProgramSpec
    from .programs import Qasm2String as Qasm2String
    from .programs import Qasm3String as Qasm3String
    from .programs import QuantumProgram as QuantumProgram
    from .programs import get_program_type_alias as get_program_type_alias
    from .programs import load_program as load_program
    from .programs import register_program_type as register_program_type
    from .programs import unregister_program_type as unregister_program_type
    from .runtime import AhsResultData as AhsResultData
    from .runtime import AhsShotResult as AhsShotResult
    from .runtime import DeviceStatus as DeviceStatus
    from .runtime import GateModelResultData as GateModelResultData
    from .runtime import JobStatus as JobStatus
    from .runtime import NoiseModel as NoiseModel
    from .runtime import QbraidClient as QbraidClient
    from .runtime import QbraidDevice as QbraidDevice
    from .runtime import QbraidJob as QbraidJob
    from .runtime import QbraidProvider as QbraidProvider
    from .runtime import QbraidSession as QbraidSession
    from .runtime import QuantumDevice as QuantumDevice
    from .runtime import QuantumJob as QuantumJob
    from .runtime import QuantumProvider as QuantumProvider
    from .runtime import Result as Result
    from .runtime import ResultData as ResultData
    from .runtime import RuntimeOptions as RuntimeOptions
    from .runtime import Session as Session
    from .runtime import TargetProfile as TargetProfile
    from .transpiler import Conversion as Conversion
    from .transpiler import ConversionGraph as ConversionGraph
    from .transpiler import ConversionScheme as ConversionScheme
    from .transpiler import translate as translate
    from .transpiler import transpile as transpile


def __getattr__(name):
    for mod_name, objects in _lazy.items():
        if name == mod_name:
            module = importlib.import_module(f".{mod_name}", __name__)
            globals()[mod_name] = module
            return module

        if name in objects:
            module = importlib.import_module(f".{mod_name}", __name__)
            obj = getattr(module, name)
            globals()[name] = obj
            return obj

    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")


def __dir__():
    return sorted(
        __all__ + list(_lazy.keys()) + [item for sublist in _lazy.values() for item in sublist]
    )



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

# pylint: disable=redefined-outer-name

"""
Unit tests for submissions to QuEra Aquila device through qBraid native runtime.

"""
import json
from typing import Any
from unittest.mock import patch

import numpy as np
import pytest
from braket.ahs.analog_hamiltonian_simulation import AnalogHamiltonianSimulation

from qbraid.programs import ExperimentType
from qbraid.runtime import AhsResultData, Result, TargetProfile
from qbraid.runtime.native import QbraidDevice, QbraidJob, QbraidProvider

from .._resources import JOB_DATA_AQUILA, RESULTS_DATA_AQUILA


@pytest.fixture
def job_data() -> dict[str, Any]:
    """Dictionary of mock QuEra Aquila job data."""
    return JOB_DATA_AQUILA.copy()


@pytest.fixture
def result_data() -> dict[str, Any]:
    """Dictionary of mock QuEra Aquila result data."""
    return RESULTS_DATA_AQUILA.copy()


@pytest.fixture
def device_id(device_data_aquila) -> str:
    """qBraid ID for QuEra Aquila device."""
    return device_data_aquila["qbraid_id"]


@pytest.fixture
def mock_job_id(job_data) -> str:
    """Mock qBraid ID for QuEra Aquila job."""
    return job_data["qbraidJobId"]


@pytest.fixture
def mock_profile(device_id, device_data_aquila) -> TargetProfile:
    """Mock QuEra Aquila TargetProfile for testing."""
    return TargetProfile(
        device_id=device_id,
        simulator=False,
        experiment_type=ExperimentType.AHS,
        num_qubits=device_data_aquila["numberQubits"],
        program_spec=QbraidProvider._get_program_spec("braket_ahs", device_id),
    )


@pytest.fixture
def mock_device(mock_profile, mock_client) -> QbraidDevice:
    """Mock QuEra Aquila QbraidDevice for testing"""
    return QbraidDevice(profile=mock_profile, client=mock_client)


@pytest.fixture
def mock_job(mock_job_id, mock_device, mock_client) -> QbraidJob:
    """Mock QuEra Aquila QbraidJob for testing."""
    return QbraidJob(job_id=mock_job_id, device=mock_device, client=mock_client)


@pytest.fixture
def mock_result_data(result_data) -> AhsResultData:
    """Mock QuEra Aquila AhsResultData for testing."""
    return AhsResultData.from_dict(result_data)


@pytest.fixture
def mock_result(device_id, mock_job_id, mock_result_data) -> Result:
    """Mock QuEra Aquila Result for testing."""
    return Result(device_id=device_id, job_id=mock_job_id, success=True, data=mock_result_data)


def test_get_aquila_device(device_id, mock_provider):
    """Test getting QuEra Aquila device."""
    device = mock_provider.get_device(device_id)
    assert device.id == device_id


def test_prepare_ahs_program(mock_device, braket_ahs, ahs_dict):
    """Test conversion of AHS program to IR."""
    assert mock_device.prepare(braket_ahs) == {"ahs": json.dumps(ahs_dict)}


@pytest.mark.filterwarnings("ignore:Device is not online*:UserWarning")
def test_submit_ahs_job_to_aquila(braket_ahs, mock_device, mock_job_id):
    """Test submitting AHS job to QuEra Aquila device."""
    job = mock_device.run(braket_ahs)
    assert job.id == mock_job_id


def test_get_aquila_job_result(mock_job, mock_result):
    """Test getting QuEra Aquila job result."""
    result = mock_job.result()
    assert result.data.get_counts() == mock_result.data.get_counts()
    assert result.data.measurements == mock_result.data.measurements


@pytest.fixture
def bloqade_program():
    """Create a Bloqade program batch."""
    try:
        # pylint: disable=import-outside-toplevel
        from bloqade.analog import var  # type: ignore
        from bloqade.analog.atom_arrangement import Square  # type: ignore

        # pylint: enable=import-outside-toplevel

        adiabatic_durations = [0.4, 3.2, 0.4]

        max_detuning = var("max_detuning")
        adiabatic_program = (
            Square(3, lattice_spacing="lattice_spacing")
            .rydberg.rabi.amplitude.uniform.piecewise_linear(
                durations=adiabatic_durations, values=[0.0, "max_rabi", "max_rabi", 0.0]
            )
            .detuning.uniform.piecewise_linear(
                durations=adiabatic_durations,
                values=[
                    -max_detuning,
                    -max_detuning,
                    max_detuning,
                    max_detuning,
                ],
            )
            .assign(max_rabi=15.8, max_detuning=16.33)
            .batch_assign(lattice_spacing=np.arange(4.0, 7.0, 1.0))
        )

        return adiabatic_program
    except ImportError as err:
        pytest.skip(f"Bloqade is not installed: {err}")

        return None


def test_device_validate_calls_for_bloqade_run_input(bloqade_program, mock_device: QbraidDevice):
    """Test that validate is called once with a list of AnalogHamiltonianSimulation."""
    mock_device.set_options(validate=0)

    with patch.object(mock_device, "validate", wraps=mock_device.validate) as mock_validate:
        _ = mock_device.run(bloqade_program, shots=10)

        mock_validate.assert_called_once()
        args, _ = mock_validate.call_args
        assert len(args) == 1

        run_input = args[0]
        assert isinstance(run_input, list)
        assert len(run_input) == 3
        assert all(isinstance(program, AnalogHamiltonianSimulation) for program in run_input)


def test_device_to_ir_calls_for_bloqade_run_input(bloqade_program, mock_device: QbraidDevice):
    """Test that to_ir is called three times with AnalogHamiltonianSimulation."""
    mock_device.set_options(validate=0)

    with patch.object(mock_device, "prepare", wraps=mock_device.prepare) as mock_to_ir:
        _ = mock_device.run(bloqade_program, shots=10)

        assert mock_to_ir.call_count == 3
        for call_args in mock_to_ir.call_args_list:
            args, _ = call_args
            assert len(args) == 1
            assert isinstance(args[0], AnalogHamiltonianSimulation)



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Unit tests for AHS visualizations

"""

from unittest.mock import patch

import pytest

from qbraid.visualization.ahs import plot_atomic_register


def test_plot_with_correct_input():
    """Test the plotting function with valid input."""
    with (
        patch("matplotlib.pyplot.show") as mock_show,
        patch("matplotlib.pyplot.savefig") as mock_save,
    ):
        plot_atomic_register([(1.0, 2.0)], [True], show=True)
        mock_show.assert_called_once()
        mock_save.assert_not_called()


def test_plot_with_save_path():
    """Test if the plot is saved to the correct path when specified."""
    with patch("matplotlib.pyplot.savefig") as mock_save:
        plot_atomic_register([(1.0, 2.0)], [True], save_path="test.png", show=False)
        mock_save.assert_called_once_with("test.png")


def test_plot_without_show():
    """Test that the plot is not shown when 'show' is False."""
    with patch("matplotlib.pyplot.show") as mock_show:
        plot_atomic_register([(1.0, 2.0)], [True], show=False)
        mock_show.assert_not_called()


def test_input_length_mismatch():
    """Test handling of input length mismatch between sites and filling."""
    with pytest.raises(ValueError) as excinfo:
        plot_atomic_register([(1.0, 2.0)], [True, False])
    assert "sites and filling must be of equal length" in str(excinfo.value)


def test_custom_labels():
    """Test that custom labels are set correctly."""
    with (
        patch("matplotlib.pyplot.xlabel") as mock_xlabel,
        patch("matplotlib.pyplot.ylabel") as mock_ylabel,
        patch("matplotlib.pyplot.title") as mock_title,
    ):
        plot_atomic_register(
            [(1.0, 2.0)],
            [True],
            title="Custom Title",
            xlabel="Custom X",
            ylabel="Custom Y",
            show=False,
        )
        mock_xlabel.assert_called_once_with("Custom X", fontsize=14)
        mock_ylabel.assert_called_once_with("Custom Y", fontsize=14)
        mock_title.assert_called_once_with("Custom Title", fontsize=16)



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Unit tests for qbraid.programs.qiskit.QiskitCircuit

"""
import pytest
from qiskit import QuantumCircuit

from qbraid.programs.exceptions import ProgramTypeError
from qbraid.programs.gate_model.qiskit import QiskitCircuit


def test_reverse_qubit_order():
    """Test reversing ordering of qubits in qiskit circuit"""
    circ = QuantumCircuit(3)
    circ.h(0)
    circ.cx(0, 2)

    qprogram = QiskitCircuit(circ)
    qprogram.reverse_qubit_order()
    reversed_circ = qprogram.program

    expected_circ = QuantumCircuit(3)
    expected_circ.h(2)
    expected_circ.cx(2, 0)

    assert (
        reversed_circ == expected_circ
    ), "The reversed circuit does not match the expected output."


def test_remove_idle_qubits_qiskit():
    """Test convert_to_contigious on qiskit circuit"""
    circuit = QuantumCircuit(3)
    circuit.h(0)
    circuit.cx(0, 1)
    qprogram = QiskitCircuit(circuit)
    qprogram.remove_idle_qubits()
    contig_circuit = qprogram.program
    assert contig_circuit.num_qubits == 2


def test_raise_program_type_error():
    """Test raising ProgramTypeError"""
    with pytest.raises(ProgramTypeError):
        QiskitCircuit("OPENQASM 2.0;qreg q[2];h q[0];cx q[0],q[1];")


def test_circuit_properties():
    """Test properties of QiskitCircuit"""
    circuit = QuantumCircuit(2)
    circuit.h(0)
    circuit.cx(0, 1)
    qprogram = QiskitCircuit(circuit)
    assert len(qprogram.qubits) == 2
    assert qprogram.num_qubits == 2
    assert qprogram.num_clbits == 0
    assert qprogram.depth == 2



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module for conversions from QASM 2 to Cirq Circuits

"""
from __future__ import annotations

from typing import TYPE_CHECKING

import pyqasm
from qbraid_core._import import LazyLoader

from qbraid._logging import logger
from qbraid.programs.exceptions import QasmError
from qbraid.transpiler.annotations import weight

cirq_qasm_import = LazyLoader("cirq_contrib", globals(), "cirq.contrib.qasm_import")
cirq_qasm_parser = LazyLoader(
    "cirq_qasm_parser", globals(), "qbraid.transpiler.conversions.qasm2.cirq_qasm_parser"
)

if TYPE_CHECKING:
    import cirq

    from qbraid.programs.typer import Qasm2StringType
    from qbraid.transpiler.conversions.qasm2.cirq_qasm_parser import QasmParser


@weight(1)
def qasm2_to_cirq(qasm: Qasm2StringType) -> cirq.Circuit:
    """Returns a Cirq circuit equivalent to the input QASM string.

    Args:
        qasm: QASM string to convert to a Cirq circuit.

    Returns:
        Cirq circuit representation equivalent to the input QASM string.
    """
    try:
        qasm_module = pyqasm.loads(qasm)
        qasm_module.unroll(external_gates=["rzz"])
        if qasm_module.has_barriers():
            logger.warning(
                "Barriers are not supported in Cirq, "
                "and will be removed during program conversion."
            )
            qasm_module.remove_barriers()
        parser: QasmParser = cirq_qasm_parser.QasmParser()
        qasm_compat = pyqasm.dumps(qasm_module)
        qasm_parsed = parser.parse(qasm_compat)
        return qasm_parsed.circuit
    except cirq_qasm_import.QasmException as err:
        raise QasmError(err) from err



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module defining BraketGateModelResultBuilder Class

"""
from __future__ import annotations

from collections import Counter
from typing import TYPE_CHECKING, Optional

import numpy as np

from qbraid.runtime.exceptions import QbraidRuntimeError
from qbraid.runtime.result_data import AhsShotResult

if TYPE_CHECKING:
    from braket.tasks.analog_hamiltonian_simulation_quantum_task_result import (
        AnalogHamiltonianSimulationQuantumTaskResult,
    )
    from braket.tasks.gate_model_quantum_task_result import GateModelQuantumTaskResult


class ResultDecodingError(QbraidRuntimeError):
    """Exception raised for errors that occur during the decoding of result data."""


class BraketGateModelResultBuilder:
    """Wrapper class for Amazon Braket result objects."""

    def __init__(
        self,
        result: GateModelQuantumTaskResult,
        partial_measurement_qubits: Optional[list[int]] = None,
    ):
        """
        Result builder with partial measurement support.

        Args:
            result: The Braket quantum task result containing measurement data.
            partial_measurement_qubits: Optional list of indices indicating which qubits
                in the measurement results correspond to the original
                partial measurements. If None, all measurements are used.
        """
        self._result = result
        self.partial_measurement_qubits = partial_measurement_qubits

    def measurements(self) -> np.ndarray:
        """
        Get measurement results as a 2D array with partial measurement support.

        Returns a 2D array where each row represents a shot and each column represents
        a qubit measurement result. If partial measurements were used, only the results
        for the originally measured qubits are returned.

        Returns:
            2D numpy array of measurement results. If partial measurements were used,
            the array is filtered to include only the originally measured qubits.
            The qubit order is reversed to match qBraid conventions.
        """
        result: GateModelQuantumTaskResult = self._result
        measurements = result.measurements

        if self.partial_measurement_qubits:
            measurements = marginal_measurement(measurements, self.partial_measurement_qubits)

        # Reverse qubit order to match qBraid conventions
        return np.flip(measurements, 1)

    def get_counts(self) -> dict[str, int]:
        """
        Get measurement counts histogram with partial measurement support.

        Returns:
            Dictionary mapping bitstrings (e.g., "101") to their counts. The bitstring
            order is reversed to match qBraid conventions. If partial measurements were
            used, only the counts for the originally measured qubits are returned.
        """
        result: GateModelQuantumTaskResult = self._result
        braket_counts = dict(result.measurement_counts)

        # Tracing out qubits if partial measurement qubits is specified
        if self.partial_measurement_qubits:
            braket_counts = marginal_count(braket_counts, self.partial_measurement_qubits)

        # Convert to qBraid format with reversed bitstring order
        qbraid_counts = {}
        for key in braket_counts:
            str_key = "".join(reversed([str(i) for i in key]))
            qbraid_counts[str_key] = braket_counts[key]
        return qbraid_counts


class BraketAhsResultBuilder:
    """Result from an Analog Hamiltonian Simulation (AHS)."""

    def __init__(
        self,
        result: AnalogHamiltonianSimulationQuantumTaskResult,
        partial_measurement_qubits: Optional[list[int]] = None,
    ):
        """
        AHS result builder.

        Args:
            result: The Braket AHS quantum task result containing measurement data.
            partial_measurement_qubits: Optional list of partial measurement qubit indices.
                Currently not used for AHS results but maintained for interface consistency.
        """
        self._result = result
        self.partial_measurement_qubits = partial_measurement_qubits

    def measurements(self) -> list[AhsShotResult]:
        """Get the list of shot results from the AHS job."""
        result: AnalogHamiltonianSimulationQuantumTaskResult = self._result
        return [
            AhsShotResult(
                success=m.status.name == "SUCCESS",
                pre_sequence=m.pre_sequence,
                post_sequence=m.post_sequence,
            )
            for m in result.measurements
        ]

    def get_counts(self) -> Optional[dict[str, int]]:
        """
        Aggregate state counts from AHS shot results.

        This function decodes the state of atoms at different sites based on
        pre- and post-measurement sequences. Each atom can be in one of three states:
        empty site ('e'), Rydberg state ('r'), or ground state ('g').

        Returns:
            Optional[dict[str, int]]: A dictionary mapping each unique state configuration to
                                      the count of its occurrences across all successful shots,
                                      or None if there are no successful measurements.

        Raises:
            ResultDecodingError: If there is an error accessing required attributes from
                the result object.

        """
        state_counts = Counter()
        states = ["e", "r", "g"]
        try:
            for shot in self.measurements():
                if shot.success:
                    pre = shot.pre_sequence
                    post = shot.post_sequence
                    # converting presequence and postsequence measurements to state_idx
                    state_idx = [
                        0 if pre_i == 0 else 1 if post_i == 0 else 2
                        for pre_i, post_i in zip(pre, post)
                    ]
                    state = "".join(states[s_idx] for s_idx in state_idx)
                    state_counts.update([state])
        except AttributeError as err:
            raise ResultDecodingError from err

        return None if not state_counts else dict(state_counts)


def marginal_measurement(
    measurements: list[list[int]], qubit_indices: list[int]
) -> list[list[int]]:
    """
    Extract marginal measurement results for the specified qubits.

    This function filters the measurement results to include only the qubits
    that were originally measured in partial measurement scenarios.

    Args:
        measurements: Raw measurement results from each shot.
            Each inner list contains the measurement result for all qubits.
        qubit_indices: List of qubit indices to keep in the results.
            Index 0 corresponds to the first element in each shot list.

    Returns:
        Filtered measurement results containing only the specified qubits.

    Example:
        >>> measurements = [[0, 1, 0, 1], [1, 0, 1, 0]]
        >>> qubit_indices = [0, 2]
        >>> marginal_measurement(measurements, qubit_indices)
        [[0, 0], [1, 1]]
    """
    return [[shot[i] for i in qubit_indices] for shot in measurements]


def marginal_count(count_dict: dict[str, int], qubit_indices: list[int]) -> dict[str, int]:
    """
    Compute marginal counts for specified qubits from measurement count data.

    This function aggregates measurement counts by summing over the unmeasured qubits,
    effectively marginalizing the probability distribution to include only the
    originally measured qubits in partial measurement scenarios.

    Args:
        count_dict: Dictionary mapping bitstrings to their occurrence counts.
            Keys are bitstrings like "0101", values are integer counts.
        qubit_indices: List of qubit indices to keep in the marginalized results.
            Index 0 corresponds to the leftmost bit in the bitstring.

    Returns:
        Dictionary mapping reduced bitstrings to their marginal counts.

    Example:
        >>> count_dict = {"0101": 10, "0111": 5, "1101": 3}
        >>> qubit_indices = [0, 2]
        >>> marginal_count(count_dict, qubit_indices)
        {"00": 10, "01": 5, "10": 3}
    """
    marginal: dict[str, int] = {}

    for bitstring, count in count_dict.items():
        reduced_bits = "".join(bitstring[i] for i in qubit_indices)
        marginal[reduced_bits] = marginal.get(reduced_bits, 0) + count

    return marginal



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module defining OpenQasm2Program class.

"""
from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
import pyqasm
from qbraid_core._import import LazyLoader

from qbraid.passes.qasm import normalize_qasm_gate_params, rebase
from qbraid.programs.exceptions import ProgramTypeError
from qbraid.programs.typer import Qasm2String, Qasm2StringType

from ._model import GateModelProgram

if TYPE_CHECKING:
    import qbraid.runtime

transpiler = LazyLoader("transpiler", globals(), "qbraid.transpiler")


class OpenQasm2Program(GateModelProgram):
    """Wrapper class for OpenQASM 2 strings."""

    def __init__(self, program: Qasm2StringType):
        super().__init__(program)
        if not isinstance(program, Qasm2String):
            raise ProgramTypeError(message=f"Expected 'str' object, got '{type(program)}'.")

        self._module = pyqasm.loads(program)

    @property
    def qubits(self) -> dict[str, int]:
        """Return the qubits acted upon by the operations in this circuit"""
        return self._module._qubit_registers

    @property
    def module(self) -> pyqasm.Module:
        """Return the pyqasm module."""
        return self._module

    @property
    def num_qubits(self) -> int:
        """Return the number of qubits in the circuit."""
        return self._module.num_qubits

    @property
    def num_clbits(self) -> int:
        """Return the number of classical bits in the circuit."""
        return self._module.num_clbits

    @property
    def depth(self) -> int:
        """Return the circuit depth (i.e., length of critical path)."""
        return self._module.depth(decompose_native_gates=False)

    def _unitary(self) -> np.ndarray:
        """Return the unitary of the QASM"""
        raise NotImplementedError

    def validate(self) -> None:
        """Validate the QASM."""
        self._module.validate()

    def transform(self, device: qbraid.runtime.QuantumDevice, **kwargs) -> None:
        """Transform program to according to device target profile."""
        if device.id == "quera_qasm_simulator":
            self._module.unroll()
            self._module.remove_measurements()
            self._program = pyqasm.dumps(self._module)

        basis_gates = device.profile.get("basis_gates")

        if basis_gates is not None and len(basis_gates) > 0:
            transformed_qasm = rebase(self.program, basis_gates, **kwargs)
            self._program = normalize_qasm_gate_params(transformed_qasm)

    def serialize(self) -> dict[str, str]:
        """Return the program in a format suitable for submission to the qBraid API."""
        return {"openQasm": pyqasm.dumps(self.module)}



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module containing quantum circuit wrapper classes providing uniform
suite of methods and functionality for supported program types.

.. currentmodule:: qbraid.programs

.. _programs_data_type:

Data Types
-----------

.. data:: QPROGRAM_REGISTRY
   :type: dict[str, Type[Any]]
   :annotation: = Maps aliases of quantum program types to their respective Python type objects.

.. autosummary::
   :toctree: ../stubs/

    QbraidMetaType
    IonQDict
    QuboCoefficientsDict
    Qasm2String
    Qasm3String
    ExperimentType

Functions
----------

.. autosummary::
   :toctree: ../stubs/

    load_program
    get_program_type_alias
    get_qasm_type_alias
    register_program_type
    unregister_program_type

Classes
--------

.. autosummary::
   :toctree: ../stubs/

    ProgramSpec
    QuantumProgram

Exceptions
-----------

.. autosummary::
   :toctree: ../stubs/

    PackageValueError
    ProgramTypeError
    ValidationError
    QasmError
    TransformError
    ProgramLoaderError

"""
import importlib
from typing import TYPE_CHECKING

from ._import import NATIVE_REGISTRY
from .alias_manager import get_program_type_alias, get_qasm_type_alias
from .exceptions import (
    PackageValueError,
    ProgramTypeError,
    QasmError,
    TransformError,
    ValidationError,
)
from .experiment import ExperimentType
from .loader import ProgramLoaderError, load_program
from .program import QuantumProgram
from .registry import (
    QPROGRAM,
    QPROGRAM_ALIASES,
    QPROGRAM_NATIVE,
    QPROGRAM_REGISTRY,
    QPROGRAM_TYPES,
    derive_program_type_alias,
    register_program_type,
    unregister_program_type,
)
from .spec import ProgramSpec
from .typer import (
    IonQDict,
    Qasm2String,
    Qasm2StringType,
    Qasm3String,
    Qasm3StringType,
    QbraidMetaType,
    QuboCoefficientsDict,
)

__all__ = [
    "NATIVE_REGISTRY",
    "PackageValueError",
    "ProgramSpec",
    "ProgramTypeError",
    "TransformError",
    "ProgramLoaderError",
    "QPROGRAM",
    "QPROGRAM_ALIASES",
    "QPROGRAM_NATIVE",
    "QPROGRAM_REGISTRY",
    "QPROGRAM_TYPES",
    "QasmError",
    "QuantumProgram",
    "derive_program_type_alias",
    "get_program_type_alias",
    "load_program",
    "get_qasm_type_alias",
    "register_program_type",
    "unregister_program_type",
    "ValidationError",
    "QbraidMetaType",
    "IonQDict",
    "QuboCoefficientsDict",
    "Qasm2String",
    "Qasm3String",
    "Qasm2StringType",
    "Qasm3StringType",
    "ExperimentType",
]

_lazy = {
    "gate_model": ["GateModelProgram"],
    "ahs": ["AnalogHamiltonianProgram", "AHSEncoder"],
    "annealing": ["AnnealingProgram", "ProblemEncoder", "ProblemType", "Problem", "QuboProblem"],
}

if TYPE_CHECKING:
    from .ahs import AHSEncoder as AHSEncoder
    from .ahs import AnalogHamiltonianProgram as AnalogHamiltonianProgram
    from .annealing import AnnealingProgram as AnnealingProgram
    from .annealing import Problem as Problem
    from .annealing import ProblemEncoder as ProblemEncoder
    from .annealing import ProblemType as ProblemType
    from .annealing import QuboProblem as QuboProblem
    from .gate_model import GateModelProgram as GateModelProgram


def __getattr__(name):
    for mod_name, objects in _lazy.items():
        if name == mod_name:
            module = importlib.import_module(f".{mod_name}", __name__)
            globals()[mod_name] = module
            return module

        if name in objects:
            module = importlib.import_module(f".{mod_name}", __name__)
            obj = getattr(module, name)
            globals()[name] = obj
            return obj

    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")


def __dir__():
    return sorted(
        __all__ + list(_lazy.keys()) + [item for sublist in _lazy.values() for item in sublist]
    )



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

# pylint: disable=redefined-outer-name

"""
Unit tests for QASM transform to basic gates

"""

from unittest.mock import MagicMock

import pytest
from openqasm3 import ast
from openqasm3.parser import parse

from qbraid.passes.exceptions import CompilationError, QasmDecompositionError
from qbraid.passes.qasm.compat import normalize_qasm_gate_params
from qbraid.passes.qasm.decompose import assert_gates_in_basis, rebase
from qbraid.programs.gate_model.qasm3 import OpenQasm3Program


# move to pyqasm
@pytest.mark.parametrize(
    "original_program, expected_program",
    [
        (
            """
OPENQASM 3.0;
include "stdgates.inc";
qubit[2] q;
crx(pi) q[0], q[1];
""",
            """OPENQASM 3.0;
include "stdgates.inc";
qubit[2] q;
rz(pi / 2) q[1];
ry(pi / 2) q[1];
cx q[0], q[1];
ry(-pi / 2) q[1];
cx q[0], q[1];
rz(-pi / 2) q[1];
""",
        ),
        (
            """
OPENQASM 3.0;
include "stdgates.inc";
qubit[2] q;
cry(pi/4) q[0], q[1];
""",
            """OPENQASM 3.0;
include "stdgates.inc";
qubit[2] q;
ry(pi / 4 / 2) q[1];
cx q[0], q[1];
ry(-(pi / 4 / 2)) q[1];
cx q[0], q[1];
""",
        ),
        (
            """
OPENQASM 3.0;
include "stdgates.inc";
qubit[2] q;
crz(pi/4) q[0], q[1];
""",
            """OPENQASM 3.0;
include "stdgates.inc";
qubit[2] q;
rz(pi / 4 / 2) q[1];
cx q[0], q[1];
rz(-(pi / 4 / 2)) q[1];
cx q[0], q[1];
""",
        ),
        (
            """OPENQASM 3.0;
include "stdgates.inc";
qubit[2] q;
cy q[0], q[1];""",
            """OPENQASM 3.0;
include "stdgates.inc";
qubit[2] q;
ry(pi / 2) q[1];
cx q[0], q[1];
ry(-(pi / 2)) q[1];
cx q[0], q[1];
s q[0];
""",
        ),
        (
            """OPENQASM 3.0;
include "stdgates.inc";
qubit[2] q;
cz q[0], q[1];""",
            """OPENQASM 3.0;
include "stdgates.inc";
qubit[2] q;
rz(pi / 2) q[1];
cx q[0], q[1];
rz(-(pi / 2)) q[1];
cx q[0], q[1];
s q[0];
""",
        ),
    ],
)
def test_rebase_controlled_rotation_gates(original_program: str, expected_program: str):
    """Test conversion of QASM3 program to basis gates"""
    converted_program = rebase(original_program, {"rz", "ry", "cx", "s"})
    assert converted_program == expected_program


# move to pyqasm
def test_rebase_bad_program_raises_decomposition_error():
    """Test conversion of QASM3 program to basis gates"""
    bad_program = """
OPENQASM 3.0;
include "stdgates.inc";
qubit[2] q;
crx() q[0], q[1];
"""
    with pytest.raises(QasmDecompositionError):
        rebase(bad_program, {"rz", "ry", "cx"})


# move to pyqasm
def test_rebase_bad_program_parser_error():
    """Test conversion of QASM3 program to basis gates"""
    bad_program = """
OPENQASM 3.0;
include "stdgates.inc";
qubit[2] q;
crx :-) q[0], q[1];
"""
    with pytest.raises(ValueError, match="Invalid OpenQASM program."):
        rebase(bad_program, "any")


@pytest.fixture
def qasm_crx_program() -> str:
    """Return a QASM3 program with a CRX gate."""
    return """
OPENQASM 3.0;
include "stdgates.inc";
qubit[2] q;
crx(pi) q[0], q[1];
"""


@pytest.fixture
def qasm_crx_parsed(qasm_crx_program) -> ast.Program:
    """Return the parsed QASM3 program with a CRX gate."""
    return parse(qasm_crx_program)


@pytest.fixture
def qasm_crx_decomposed() -> str:
    """Return a QASM3 program with a CRX gate decomposed into basis gates."""
    return """
OPENQASM 3.0;
include "stdgates.inc";
qubit[2] q;
rz(pi / 2) q[1];
ry(pi / 2) q[1];
cx q[0], q[1];
ry(-pi / 2) q[1];
cx q[0], q[1];
rz(-pi / 2) q[1];
"""


# move to pyqasm
def test_rebase_raises_for_basis_gates_type_error(qasm_crx_program):
    """Test rebase raises a TypeError if basis_gates is not a set"""
    with pytest.raises(
        TypeError, match="Basis gate set must be a set of strings or a string identifier."
    ):
        rebase(qasm_crx_program, 42)


# move to pyqasm
def test_rebase_raises_for_bad_basis_gate_set_identifier(qasm_crx_program):
    """Test rebase raises a TypeError if gateset is not a set"""
    with pytest.raises(ValueError, match="Invalid basis gate set identifier."):
        rebase(qasm_crx_program, gateset="null")


# move to pyqasm
def test_assert_gates_in_basis(qasm_crx_parsed: ast.Program):
    """Test assertion that all gates are in the basis"""
    with pytest.raises(
        ValueError, match="OpenQASM program uses gate 'crx' which is not in the basis gate set."
    ):
        assert_gates_in_basis(qasm_crx_parsed, {"h"})


# move to pyqasm
def test_raise_value_error_on_empty_basis_set(qasm_crx_parsed: ast.Program):
    """Test raising a ValueError when the basis set is empty"""
    with pytest.raises(ValueError, match="Basis gate set cannot be empty."):
        rebase(qasm_crx_parsed, set())


# move to pyqasm
def test_rebase_program_already_in_basis_set(qasm_crx_program: str):
    """Test that a program already in the basis set is not modified"""
    converted_program = rebase(qasm_crx_program, {"crx"})
    assert converted_program.strip() == qasm_crx_program.strip()


# move to pyqasm
def test_rebase_raises_for_unsatisfied_predicates(qasm_crx_program: str):
    """Test that rebase raises an error if the predicates are not satisfied"""
    gateset = {"h"}

    with pytest.raises(
        CompilationError,
        match="Rebasing the specified quantum program to the provided "
        f"basis gate set {gateset} is not supported.",
    ):
        rebase(qasm_crx_program, gateset)


# move to pyqasm
def test_rebase_unsatisfied_predicates_no_check(qasm_crx_program: str):
    """Test that rebase raises an error if the predicates are not satisfied"""
    assert rebase(qasm_crx_program, {"h"}, require_predicates=False) == qasm_crx_program


# move to pyqasm
def test_rebase_over_supported_decomposition_basis_set(
    qasm_crx_program: str, qasm_crx_decomposed: str
):
    """Test rebasing over basis set that matches the gates used in supported decomposition"""
    rebased_program = rebase(qasm_crx_program, {"rz", "ry", "cx"})
    assert rebased_program.strip() == qasm_crx_decomposed.strip()


def test_transform_from_device_gateset(qasm_crx_program: str, qasm_crx_decomposed: str):
    """Test transforming a program to the basis gates of a device"""
    device = MagicMock()
    device.profile.get.return_value = {"rz", "ry", "cx"}
    program = OpenQasm3Program(qasm_crx_program)
    program.transform(device=device)
    expected = normalize_qasm_gate_params(qasm_crx_decomposed).strip()
    assert program.program == expected



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

# pylint: disable=redefined-outer-name

"""
Unit tests for TargetProfile class

"""
import pytest
from pydantic import ValidationError

from qbraid.programs import ExperimentType
from qbraid.programs.spec import ProgramSpec
from qbraid.runtime.profile import TargetProfile


@pytest.fixture
def valid_program_spec() -> ProgramSpec:
    """Fixture to provide a valid ProgramSpec instance."""
    return ProgramSpec(program_type=str, alias="qasm2")


def create_target_profile(experiment_type, basis_gates, program_spec):
    """Helper function to create a TargetProfile with given parameters."""
    return TargetProfile(
        device_id="device123",
        simulator=True,
        experiment_type=experiment_type,
        num_qubits=5,
        program_spec=program_spec,
        provider_name="Heisenberg",
        basis_gates=basis_gates,
    )


@pytest.fixture
def target_profile_openqasm(valid_program_spec) -> TargetProfile:
    """Fixture to provide a valid TargetProfile instance with OPENQASM action type."""
    return create_target_profile(
        experiment_type=ExperimentType.GATE_MODEL,
        basis_gates=["ms", "gpi", "gpi2"],
        program_spec=valid_program_spec,
    )


@pytest.fixture
def target_profile_ahs(valid_program_spec) -> list[TargetProfile]:
    """Fixture to provide a valid TargetProfile instance with AHS action type."""
    return create_target_profile(
        experiment_type=ExperimentType.AHS, basis_gates=None, program_spec=valid_program_spec
    )


@pytest.mark.parametrize(
    "simulator, local",
    [
        (True, False),  # simulator
        (False, False),  # QPU
        (True, True),  # local simulator
    ],
)
def test_target_profile_simulator_and_local_fields(simulator, local, valid_program_spec):
    """Ensure that the constructor correctly populates simulator and local fields."""
    profile = TargetProfile(
        device_id="device123",
        simulator=simulator,
        local=local,
        num_qubits=5,
        program_spec=valid_program_spec,
    )
    assert profile.simulator == simulator
    assert profile.get("local") == local


def test_target_profile_raise_for_bad_input():
    """
    Test that the TargetProfile constructor raises appropriate TypeError
    or ValueError for invalid inputs.
    """
    with pytest.raises(ValidationError):
        TargetProfile(device_id=123, simulator=True)
    with pytest.raises(ValidationError):
        TargetProfile(device_id="device123", simulator="invalid_type")
    with pytest.raises(ValidationError):
        TargetProfile(device_id="device123", simulator=True, experiment_type="invalid_action")
    with pytest.raises(ValidationError):
        TargetProfile(device_id="device123", simulator=True, num_qubits="five")
    with pytest.raises(ValidationError):
        TargetProfile(device_id="device123", simulator=True, program_spec="not_a_spec")
    with pytest.raises(ValidationError):
        TargetProfile(device_id="device123", simulator=True, provider_name=10)
    with pytest.raises(ValidationError):
        TargetProfile(device_id="1234", simulator=True, basis_gates=123)
    with pytest.raises(ValidationError):
        TargetProfile(device_id="1234", simulator=True, basis_gates=[1, 2, 3])
    with pytest.raises(ValidationError):
        TargetProfile(device_id="1234", simulator=True, basis_gates=[1, "cx", "h"])
    with pytest.raises(ValidationError):
        TargetProfile(
            device_id="device123",
            simulator=True,
            experiment_type=ExperimentType.GATE_MODEL,
            basis_gates=99,
        )
    with pytest.raises(ValidationError):
        TargetProfile(
            device_id="device123",
            simulator=True,
            experiment_type=ExperimentType.GATE_MODEL,
            basis_gates=[None],
        )
    with pytest.raises(ValidationError):
        TargetProfile(
            device_id="device123",
            simulator=True,
            basis_gates=["ms", "gpi", "gpi2"],
        )


def test_target_profile_openqasm_len(target_profile_openqasm):
    """Test the __len__ method of TargetProfile"""
    assert len(target_profile_openqasm) == 7


def test_target_profile_ahs_len(target_profile_ahs):
    """Test the __len__ method of TargetProfile"""
    assert len(target_profile_ahs) == 6


def test_target_profile_bad_basis_gates_raises(valid_program_spec):
    """Test raising ValueError when giving basis_gates with non-OPENQASM action type."""
    for experiment_type in [ExperimentType.AHS, None]:
        with pytest.raises(ValidationError):
            create_target_profile(
                experiment_type=experiment_type,
                basis_gates=["x", "y", "z"],
                program_spec=valid_program_spec,
            )


def test_target_profile_getitem(target_profile_openqasm):
    """Test the __getitem__ method of TargetProfile."""
    assert target_profile_openqasm["device_id"] == "device123"
    assert target_profile_openqasm["simulator"] is True
    assert target_profile_openqasm["experiment_type"] == ExperimentType.GATE_MODEL
    assert target_profile_openqasm["basis_gates"] == {"gpi", "gpi2", "ms"}


def test_duplicate_basis_gates_removed(valid_program_spec):
    """Test that duplicate basis gates are removed."""
    profile = create_target_profile(
        experiment_type=ExperimentType.GATE_MODEL,
        basis_gates=["X", "x", "x", "Y", "z"],
        program_spec=valid_program_spec,
    )
    assert profile["basis_gates"] == {"x", "y", "z"}


def test_basis_gates_none():
    """Test that passing None to basis_gates does not raise an error."""
    profile = TargetProfile(device_id="1234", simulator=True, basis_gates=None)
    assert profile.basis_gates is None
